import os
from typing import Optional

import django_countries.widgets
import htmlgenerator as hg
from _strptime import TimeRE
from django.conf import settings
from django.forms import widgets
from django.utils import formats
from django.utils.translation import gettext_lazy as _
from phonenumber_field.formfields import PhoneNumberField

from ..button import Button
from ..icon import Icon
from ..tag import Tag
from .helpers import REQUIRED_LABEL, Label, to_php_formatstr

# Missing widget implementations:
# DateTimeInput


# A plain, largely unstyled input element
# It provides the basic structure for fields used
# in bread forms
class BaseWidget(hg.DIV):
    # used to mark that this class can be used in place of the according django widget or field
    # all bread widgets must have this if they should be used automatically to render form fields
    django_widget = None

    # default attributes which are used to create the input element in a standardized way for many inputs
    carbon_input_class: str = ""
    carbon_input_error_class: str = ""
    input_type: Optional[str] = None

    # __init__ of derived classes should support the following parameters
    # label: bread.layout.components.forms.utils.Label
    # help_text: Optional[Any]
    # errors: bread.layout.components.forms.utils.ErrorList
    #         ``hg.If(getattr(errors, "condition", None), ...`` can be used
    #         to render parts depending if there are errors
    # inputelement_attrs: Union[Lazy[dict], dict],
    # boundfield: Optional[django.forms.BoundField],

    def __init__(self, *args, **kwargs):
        # prevent rendering of any of the special kwargs we use in generate_formfield
        for param in [
            "label_element",
            "label",
            "help_text_element",
            "error_element",
            "inputelement_attrs",
            "boundfield",
        ]:
            if param in kwargs:
                raise Exception(
                    f"Widget {type(self)} tries to render parameter {param}"
                )
        super().__init__(*args, **kwargs)

    def get_input_element(self, inputelement_attrs, errors, **kwargs):
        return hg.INPUT(
            type=self.input_type,
            lazy_attributes=_combine_lazy_dict(
                _append_classes(
                    inputelement_attrs or {},
                    self.carbon_input_class,
                    hg.If(
                        getattr(errors, "condition", False),
                        self.carbon_input_error_class,
                    ),
                ),
                kwargs,
            ),
            data_invalid=hg.If(getattr(errors, "condition", False), True),
        )

    def with_fieldwrapper(self):
        return hg.DIV(self, _class="bx--form-item")


class HiddenInput(BaseWidget):
    django_widget = widgets.HiddenInput
    input_type = "hidden"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        super().__init__(self.get_input_element(inputelement_attrs, errors))


class TextInput(BaseWidget):
    django_widget = widgets.TextInput
    carbon_input_class = "bx--text-input"
    carbon_input_error_class = "bx--text-input--invalid"
    input_type = "text"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        icon=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        super().__init__(
            label,
            hg.DIV(
                hg.If(
                    getattr(errors, "condition", None),
                    Icon(
                        "warning--filled",
                        size=16,
                        _class="bx--text-input__invalid-icon",
                    ),
                ),
                self.get_input_element(inputelement_attrs, errors),
                hg.If(
                    icon,
                    hg.If(
                        hg.F(lambda c: isinstance(icon, str)),
                        Icon(
                            icon,
                            size=16,
                            _class="text-input-icon",
                        ),
                        icon,
                    ),
                ),
                _class=(
                    "bx--text-input__field-wrapper"
                    + (" text-input-with-icon" if icon is not None else "")
                ),
                data_invalid=hg.If(getattr(errors, "condition", None), True),
            ),
            errors,
            help_text,
            **hg.merge_html_attrs(attributes, {"_class": "bx--text-input-wrapper"}),
        )


class PhoneNumberInput(TextInput):
    input_type = "tel"
    django_widget = PhoneNumberField

    def __init__(self, *args, **attributes):
        super().__init__(*args, icon="phone", **attributes)


class UrlInput(TextInput):
    django_widget = widgets.URLInput
    input_type = "url"

    def __init__(self, *args, **attributes):
        super().__init__(*args, icon="link", **attributes)


class EmailInput(TextInput):
    django_widget = widgets.EmailInput
    input_type = "email"

    def __init__(self, *args, **attributes):
        super().__init__(*args, icon="email", **attributes)


class NumberInput(BaseWidget):
    django_widget = widgets.NumberInput
    input_type = "number"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        super().__init__(
            label,
            hg.DIV(
                self.get_input_element(inputelement_attrs, errors),
                hg.DIV(
                    hg.BUTTON(
                        Icon("caret--up", size=16),
                        _class="bx--number__control-btn up-icon",
                        type="button",
                    ),
                    hg.BUTTON(
                        Icon("caret--down", size=16),
                        _class="bx--number__control-btn down-icon",
                        type="button",
                    ),
                    _class="bx--number__controls",
                ),
                _class="bx--number__input-wrapper",
            ),
            errors,
            help_text,
            data_numberinput=True,
            data_invalid=hg.If(errors.condition, True),
            **hg.merge_html_attrs(attributes, {"_class": "bx--number"}),
        )


class TimeInput(TextInput):
    django_widget = widgets.TimeInput
    input_type = "time"


class PasswordInput(TextInput):
    django_widget = widgets.PasswordInput
    carbon_input_class = "bx--password-input bx--text-input"
    carbon_input_error_class = "bx--text-input--invalid"
    input_type = "password"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        showhidebtn = Button(
            _("Show password"),
            icon=hg.BaseElement(
                Icon("view--off", _class="bx--icon--visibility-off", hidden="true"),
                Icon("view", _class="bx--icon--visibility-on"),
            ),
            notext=True,
        )
        # override attributes from button
        showhidebtn.attributes["_class"] = (
            "bx--text-input--password__visibility__toggle bx--tooltip__trigger "
            "bx--tooltip--a11y bx--tooltip--bottom bx--tooltip--align-center"
        )
        super().__init__(
            label=label,
            help_text=help_text,
            errors=errors,
            inputelement_attrs=_combine_lazy_dict(
                inputelement_attrs, {"data_toggle_password_visibility": True}
            ),
            icon=showhidebtn,
            **hg.merge_html_attrs(
                attributes,
                {"data_text_input": True, "_class": "bx--password-input-wrapper"},
            ),
        )


class Textarea(BaseWidget):
    django_widget = widgets.Textarea
    carbon_input_class = "bx--text-area bx--text-area--v2"
    carbon_input_error_class = "bx--text-area--invalid"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        super().__init__(
            label,
            hg.DIV(
                hg.If(
                    getattr(errors, "condition", None),
                    Icon(
                        "warning--filled",
                        size=16,
                        _class="bx--text-area__invalid-icon",
                    ),
                ),
                hg.TEXTAREA(
                    boundfield.value() if boundfield else None,
                    lazy_attributes=_combine_lazy_dict(
                        _append_classes(
                            inputelement_attrs or {},
                            self.carbon_input_class,
                            hg.If(
                                getattr(errors, "condition", None),
                                self.carbon_input_error_class,
                            ),
                        ),
                        {"value": None},
                    ),
                ),
                _class="bx--text-area__wrapper",
                data_invalid=hg.If(getattr(errors, "condition", None), True),
            ),
            help_text,
            errors,
            **attributes,
        )


class Select(BaseWidget):
    django_widget = widgets.Select
    carbon_input_class = "bx--select-input"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        inline=False,
        choices=None,  # for non-django-form select elements use this
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        select_wrapper = hg.DIV(
            hg.SELECT(
                hg.Iterator(
                    _optgroups_from_choices(
                        choices,
                        name=inputelement_attrs.get("name"),
                        value=inputelement_attrs.get("value"),
                    )
                    if choices
                    else _gen_optgroup(boundfield),
                    "optgroup",
                    hg.If(
                        hg.C("optgroup.0"),
                        hg.OPTGROUP(
                            hg.Iterator(
                                hg.C("optgroup.1"),
                                "option",
                                hg.OPTION(
                                    hg.C("option.label"),
                                    _class="bx--select-option",
                                    value=hg.C("option.value"),
                                    lazy_attributes=hg.C("option.attrs"),
                                ),
                            ),
                            _class="bx--select-optgroup",
                            label=hg.C("optgroup.0"),
                        ),
                        hg.Iterator(
                            hg.C("optgroup.1"),
                            "option",
                            hg.OPTION(
                                hg.C("option.label"),
                                _class="bx--select-option",
                                value=hg.C("option.value"),
                                lazy_attributes=hg.C("option.attrs"),
                            ),
                        ),
                    ),
                ),
                lazy_attributes=_append_classes(
                    inputelement_attrs,
                    self.carbon_input_class,
                    hg.If(
                        getattr(errors, "condition", None),
                        self.carbon_input_error_class,
                    ),
                ),
            ),
            Icon(
                "chevron--down",
                size=16,
                _class="bx--select__arrow",
                aria_hidden="true",
            ),
            hg.If(
                getattr(errors, "condition", None),
                Icon(
                    "warning--filled",
                    size=16,
                    _class="bx--select__invalid-icon",
                ),
            ),
            _class="bx--select-input__wrapper",
            data_invalid=hg.If(getattr(errors, "condition", None), True),
        )
        super().__init__(
            label,
            hg.If(
                inline,
                hg.DIV(
                    select_wrapper,
                    errors,
                    _class="bx--select-input--inline__wrapper",
                ),
                select_wrapper,
            ),
            help_text,
            hg.If(inline, None, errors),  # not displayed if this is inline
            **hg.merge_html_attrs(
                attributes,
                {
                    "_class": hg.BaseElement(
                        "bx--select",
                        hg.If(inline, " bx--select--inline"),
                        hg.If(
                            getattr(errors, "condition", None), " bx--select--invalid"
                        ),
                        hg.If(
                            inputelement_attrs.get("disabled"),
                            " bx--select--disabled",
                        ),
                    ),
                },
            ),
        )


class NullBooleanSelect(Select):
    django_widget = widgets.NullBooleanSelect


class SelectMultiple(BaseWidget):
    django_widget = widgets.SelectMultiple

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,  # for django-form select elements use this
        choices=None,  # for non-django-form select elements use this
        **attributes,  # for non-django-form select elements use this
    ):
        inputelement_attrs = inputelement_attrs or {}
        optgroups = (
            _optgroups_from_choices(
                choices,
                name=inputelement_attrs.get("name"),
                value=inputelement_attrs.get("value"),
            )
            if choices
            else _gen_optgroup(boundfield)
        )

        def countselected(context):
            options = [o for og in hg.resolve_lazy(optgroups, context) for o in og[1]]
            return len([o for o in options if o and o["selected"]])

        searchfieldid = hg.html_id(self)
        super().__init__(
            label,
            hg.If(
                inputelement_attrs.get("disabled"),
                hg.DIV(
                    hg.Iterator(
                        optgroups,
                        "optiongroup",
                        hg.Iterator(
                            hg.C("optiongroup.1"),
                            "option",
                            hg.If(hg.C("option.selected"), Tag(hg.C("option.label"))),
                        ),
                    )
                ),
                hg.DIV(
                    hg.DIV(
                        hg.DIV(
                            hg.F(countselected),
                            Icon(
                                "close",
                                focusable="false",
                                size=15,
                                role="img",
                                onclick="clearMultiselect(this.parentElement.parentElement.parentElement)",
                            ),
                            role="button",
                            _class="bx--list-box__selection bx--list-box__selection--multi bx--tag--filter",
                            tabindex="0",
                            title="Clear all selected items",
                        ),
                        hg.INPUT(
                            id=searchfieldid,
                            _class="bx--text-input",
                            placeholder="Filter...",
                            onclick="this.parentElement.nextElementSibling.style.display = 'block'",
                            onkeyup="filterOptions(this.parentElement.parentElement)",
                        ),
                        hg.DIV(
                            Icon(
                                "chevron--down", size=16, role="img", focusable="false"
                            ),
                            _class="bx--list-box__menu-icon",
                            onclick="this.parentElement.nextElementSibling.style.display = this.parentElement.nextElementSibling.style.display == 'none' ? 'block' : 'none';",
                        ),
                        role="button",
                        _class="bx--list-box__field",
                        tabindex="0",
                        onload="window.addEventListener('click', (e) => {this.nextElementSibling.style.display = 'none'})",
                    ),
                    hg.FIELDSET(
                        hg.Iterator(
                            optgroups,
                            "optgroup",
                            hg.Iterator(
                                hg.C("optgroup.1"),
                                "option",
                                hg.DIV(
                                    hg.DIV(
                                        hg.DIV(
                                            hg.LABEL(
                                                hg.INPUT(
                                                    type="checkbox",
                                                    readonly=True,
                                                    _class="bx--checkbox",
                                                    value=hg.C("option.value"),
                                                    lazy_attributes=hg.C(
                                                        "option.attrs"
                                                    ),
                                                    onchange="updateMultiselect(this.closest('.bx--multi-select'))",
                                                    checked=hg.C("option.selected"),
                                                    name=hg.C("option.name"),
                                                ),
                                                hg.SPAN(
                                                    _class="bx--checkbox-appearance"
                                                ),
                                                hg.SPAN(
                                                    hg.C("option.label"),
                                                    _class="bx--checkbox-label-text",
                                                ),
                                                title=hg.C("option.label"),
                                                _class="bx--checkbox-label",
                                            ),
                                            _class="bx--form-item bx--checkbox-wrapper",
                                        ),
                                        _class="bx--list-box__menu-item__option",
                                    ),
                                    _class="bx--list-box__menu-item",
                                ),
                            ),
                        ),
                        _class="bx--list-box__menu",
                        role="listbox",
                        style="display: none",
                    ),
                    _class=hg.BaseElement(
                        "bx--multi-select bx--list-box bx--multi-select--selected bx--combo-box bx--multi-select--filterable",
                        hg.If(
                            inputelement_attrs.get("disabled"),
                            " bx--list-box--disabled",
                        ),
                    ),
                    data_invalid=hg.If(getattr(errors, "condition", None), True),
                ),
            ),
            help_text,
            errors,
            **hg.merge_html_attrs(
                attributes,
                {
                    "onclick": "event.stopPropagation()",
                    "_class": "bx--list-box__wrapper",
                },
            ),
        )


class Checkbox(BaseWidget):
    django_widget = widgets.CheckboxInput
    carbon_input_class = "bx--checkbox"
    input_type = "checkbox"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        attrs = {}
        if boundfield:
            attrs["checked"] = hg.F(
                lambda c: hg.resolve_lazy(boundfield, c).field.widget.check_test(
                    hg.resolve_lazy(boundfield, c).value()
                )
            )
            attrs["value"] = None
        inputelement_attrs = _combine_lazy_dict(inputelement_attrs, attrs)
        # labels for checkboxes are treated a bit different, need to use plain value
        label = hg.F(
            lambda c, label=label: getattr(hg.resolve_lazy(label, c), "label", label)
        )
        required = hg.F(lambda c, label=label: hg.resolve_lazy(label, c) is not None)
        super().__init__(
            hg.LABEL(
                self.get_input_element(inputelement_attrs, errors),
                label,
                hg.If(
                    inputelement_attrs.get("required"), hg.If(required, REQUIRED_LABEL)
                ),
                _class=hg.BaseElement(
                    "bx--checkbox-label",
                    hg.If(inputelement_attrs.get("disabled"), " bx--label--disabled"),
                ),
                data_contained_checkbox_state=hg.If(
                    inputelement_attrs.get("checked"),
                    "true",
                    "false",
                ),
                data_invalid=hg.If(getattr(errors, "condition", False), True),
            ),
            help_text,
            errors,
            **hg.merge_html_attrs(attributes, {"_class": "bx--checkbox-wrapper"}),
        )


class CheckboxSelectMultiple(BaseWidget):
    django_widget = widgets.CheckboxSelectMultiple
    carbon_input_class = "bx--checkbox"
    input_type = "checkbox"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        super().__init__(
            hg.FIELDSET(
                label,
                hg.Iterator(
                    boundfield.subwidgets,
                    "checkbox",
                    Checkbox(
                        label=Label(hg.C("checkbox").data["label"]),
                        inputelement_attrs=_combine_lazy_dict(
                            _combine_lazy_dict(
                                inputelement_attrs,
                                {
                                    "name": hg.C("checkbox").data["name"],
                                    "value": hg.C("checkbox").data["value"],
                                    "checked": hg.C("checkbox").data["selected"],
                                },
                            ),
                            hg.C("checkbox").data["attrs"],
                        ),
                    ),
                ),
                data_invalid=hg.If(getattr(errors, "condition", False), True),
            ),
            help_text,
            errors,
            **attributes,
        )


class RadioButton(BaseWidget):
    django_widget = None  # only used inside RadioSelect
    carbon_input_class = "bx--radio-button"
    input_type = "radio"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        attrs = {}
        if boundfield:
            attrs["checked"] = hg.F(
                lambda c: hg.resolve_lazy(boundfield, c).field.widget.check_test(
                    hg.resolve_lazy(boundfield, c).value()
                )
            )
        inputelement_attrs = _combine_lazy_dict(inputelement_attrs, attrs)
        label = None if label is None else label.label
        super().__init__(
            self.get_input_element(inputelement_attrs, errors),
            hg.LABEL(
                hg.SPAN(_class="bx--radio-button__appearance"),
                hg.SPAN(label, _class="bx--radio-button__label-text"),
                _class="bx--radio-button__label",
                _for=inputelement_attrs.get("id"),
            ),
            help_text,
            errors,
            **hg.merge_html_attrs(attributes, {"_class": "bx--radio-button-wrapper"}),
        )


class RadioSelect(BaseWidget):
    django_widget = widgets.RadioSelect
    carbon_input_class = "bx--radio-button"
    input_type = "radio"

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        super().__init__(
            hg.FIELDSET(
                label,
                hg.DIV(
                    hg.Iterator(
                        boundfield.subwidgets,
                        "radiobutton",
                        RadioButton(
                            label=Label(hg.C("radiobutton").data["label"]),
                            inputelement_attrs=_combine_lazy_dict(
                                _combine_lazy_dict(
                                    inputelement_attrs,
                                    {
                                        "name": hg.C("radiobutton").data["name"],
                                        "value": hg.C("radiobutton").data["value"],
                                        "checked": hg.C("radiobutton").data["selected"],
                                    },
                                ),
                                hg.C("radiobutton").data["attrs"],
                            ),
                        ),
                    ),
                    _class="bx--radio-button-group  bx--radio-button-group--vertical",
                ),
                data_invalid=hg.If(getattr(errors, "condition", False), True),
            ),
            help_text,
            errors,
            **attributes,
        )


class DatePicker(BaseWidget):
    django_widget = widgets.DateInput
    carbon_input_class = "bx--date-picker__input"
    input_type = "text"  # prevent browser style date picker but use carbon design

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        style_short=False,
        style_simple=False,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}

        def format_date_value(context):
            bfield = hg.resolve_lazy(boundfield, context)
            return bfield.field.widget.format_value(bfield.value())

        super().__init__(
            hg.DIV(
                label,
                hg.If(
                    style_simple,
                    self.get_input_element(
                        inputelement_attrs,
                        errors,
                        data_invalid=hg.If(getattr(errors, "condition"), True),
                        pattern=hg.F(
                            lambda c: (
                                TimeRE()
                                .compile(
                                    hg.resolve_lazy(boundfield, c).field.widget.format
                                    or formats.get_format(
                                        hg.resolve_lazy(
                                            boundfield, c
                                        ).field.widget.format_key
                                    )[0]
                                )
                                .pattern
                            )
                        ),
                        value=hg.F(format_date_value),
                    ),
                    hg.DIV(
                        self.get_input_element(
                            inputelement_attrs,
                            errors,
                            data_date_picker_input=True,
                            data_invalid=hg.If(getattr(errors, "condition"), True),
                            data_date_format=hg.F(
                                lambda c: to_php_formatstr(
                                    hg.resolve_lazy(boundfield, c).field.widget.format,
                                    hg.resolve_lazy(
                                        boundfield, c
                                    ).field.widget.format_key,
                                )
                            ),
                            value=hg.F(format_date_value),
                        ),
                        Icon(
                            "calendar",
                            size=16,
                            _class="bx--date-picker__icon",
                            data_date_picker_icon="true",
                        ),
                        _class="bx--date-picker-input__wrapper",
                    ),
                ),
                help_text,
                errors,
                _class="bx--date-picker-container",
            ),
            **hg.merge_html_attrs(
                attributes,
                {
                    "data_date_picker": not style_simple,
                    "data_date_picker_type": None if style_simple else "single",
                    "_class": hg.BaseElement(
                        "bx--date-picker",
                        hg.If(
                            style_simple,
                            " bx--date-picker--simple",
                            " bx--date-picker--single",
                        ),
                        hg.If(style_short, " bx--date-picker--short"),
                    ),
                },
            ),
        )


class FileInput(BaseWidget):
    django_widget = widgets.FileInput
    carbon_input_class = "bx--file-input bx--visually-hidden"
    input_type = "file"
    clearable = False

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes,
    ):
        inputelement_attrs = inputelement_attrs or {}
        uploadbutton = hg.LABEL(
            hg.SPAN(_("Select file"), role="button"),
            tabindex=0,
            _class=hg.BaseElement(
                "bx--btn bx--btn--tertiary",
                hg.If(inputelement_attrs.get("disabled"), " bx--btn--disabled"),
            ),
            data_file_drop_container=True,
            disabled=inputelement_attrs.get("disabled"),
            data_invalid=hg.If(getattr(errors, "condition", False), True),
            _for=inputelement_attrs.get("id"),
        )
        input = self.get_input_element(
            inputelement_attrs,
            errors,
            onload="""
that = this;
document.addEventListener('change', (e) => {
    that.parentElement.querySelector('[data-file-container]').innerHTML = '';
    var widget = new CarbonComponents.FileUploader(that.parentElement);
    widget._displayFilenames();
    widget.setState('edit');
});
""",
        )
        # we can only clear the field if it originates form a django field
        # otherwise it has no use
        clearbox = None
        if boundfield:
            checkbox_name = hg.F(
                lambda c: hg.resolve_lazy(
                    boundfield, c
                ).field.widget.clear_checkbox_name(
                    hg.resolve_lazy(boundfield, c).html_name
                )
            )
            checkbox_id = hg.F(
                lambda c: hg.resolve_lazy(boundfield, c).field.widget.clear_checkbox_id(
                    hg.resolve_lazy(checkbox_name, c)
                )
            )
            clearbox = hg.If(
                self.clearable,
                hg.INPUT(
                    type="checkbox",
                    name=checkbox_name,
                    id=checkbox_id,
                    style="display: none",
                ),
            )

        # clearbutton is always used, to allow clearing a just selected field in the browser
        clearbutton = hg.If(
            self.clearable,
            hg.SPAN(
                hg.BUTTON(
                    Icon("close", size=16),
                    _class="bx--file-close",
                    type="button",
                    aria_label="close",
                    onclick=hg.If(
                        clearbox,
                        hg.format("$('#{}').checked = 'checked';", checkbox_id),
                    ),
                ),
                data_for=inputelement_attrs.get("id"),
                _class="bx--file__state-container",
            ),
        )

        super().__init__(
            label,
            hg.DIV(
                uploadbutton,
                input,
                clearbox,
                hg.DIV(
                    hg.If(
                        inputelement_attrs.get("value"),
                        hg.SPAN(
                            hg.P(
                                hg.If(
                                    hg.F(
                                        lambda c: hasattr(
                                            hg.resolve_lazy(inputelement_attrs, c)
                                            .get("value")
                                            .file,
                                            "name",
                                        )
                                    ),
                                    hg.A(
                                        hg.F(
                                            lambda c: os.path.basename(
                                                hg.resolve_lazy(inputelement_attrs, c)
                                                .get("value")
                                                .name
                                            )
                                        ),
                                        href=hg.F(
                                            lambda c: settings.MEDIA_URL
                                            + hg.resolve_lazy(inputelement_attrs, c)
                                            .get("value")
                                            .name
                                        ),
                                    ),
                                    hg.F(
                                        lambda c: os.path.basename(
                                            hg.resolve_lazy(inputelement_attrs, c)
                                            .get("value")
                                            .name
                                        )
                                    ),
                                ),
                                _class="bx--file-filename",
                            ),
                            clearbutton,
                            _class="bx--file__selected-file",
                        ),
                    ),
                    data_file_container=True,
                    _class="bx--file-container",
                ),
                help_text,
                errors,
                _class="bx--file",
                data_file=True,
            ),
            **attributes,
        )


class ClearableFileInput(FileInput):
    django_widget = widgets.ClearableFileInput
    clearable = True


class LazySelect(Select):
    django_widget = django_countries.widgets.LazySelect


def _append_classes(lazy_attrs, *_classes):
    def wrapper_func(context):
        _classlist = []
        for _class in _classes:
            _classlist.append(_class)
            _classlist.append(" ")
        ret = hg.resolve_lazy(lazy_attrs, context) or {}
        ret["_class"] = hg.BaseElement(ret.get("_class", ""), " ", *_classlist)
        return ret

    return hg.F(wrapper_func)


def _combine_lazy_dict(attrs1, attrs2):
    return hg.F(
        lambda c: {
            **(hg.resolve_lazy(attrs1, c) or {}),
            **(hg.resolve_lazy(attrs2, c) or {}),
        }
    )


def _gen_optgroup(boundfield):
    def wrapper(context):
        bfield = hg.resolve_lazy(boundfield, context)
        return bfield.field.widget.optgroups(
            bfield.name,
            bfield.field.widget.format_value(bfield.value()),
        )

    return hg.F(wrapper)


# based on https://github.com/django/django/blob/c6c6cd3c5ad9c36795bb120e521590424f034ae4/django/forms/widgets.py#L587
def _optgroups_from_choices(optchoices, name, value):
    groups = []

    for index, (option_value, option_label) in enumerate(optchoices):
        if option_value is None:
            option_value = ""

        subgroup = []
        if isinstance(option_label, (list, tuple)):
            group_name = option_value
            subindex = 0
            choices = option_label
        else:
            group_name = None
            subindex = None
            choices = [(option_value, option_label)]
        groups.append((group_name, subgroup, index))

        for subvalue, sublabel in choices:
            selected = hg.F(
                lambda c, v=subvalue: hg.resolve_lazy(v, c) == hg.resolve_lazy(value, c)
            )
            subgroup.append(
                {
                    "name": name,
                    "value": subvalue,
                    "label": sublabel,
                    "selected": selected,
                    "attrs": {
                        "selected": selected,
                    },
                }
            )
            if subindex is not None:
                subindex += 1
    return groups



import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from .icon import Icon


class ProgressStep(hg.LI):
    STATUS = {
        "warning": "warning",
        "current": "circle--filled",
        "complete": "checkmark--outline",
        "incomplete": "radio-button",
    }

    def __init__(
        self, label, status, optional=False, tooltip=None, disabled=False, **kwargs
    ):
        if status not in ProgressStep.STATUS:
            raise ValueError(f"{status} must be one of {ProgressStep.STATUS}")
        kwargs["_class"] = (
            kwargs.get("_class", "") + f" bx--progress-step bx--progress-step--{status}"
        )
        if disabled:
            kwargs["aria_disabled"] = "true"
            kwargs["_class"] += "  bx--progress-step--disabled"
        elements = [
            Icon(ProgressStep.STATUS[status], size=16),
            hg.P(label, tabindex=0, _class="bx--progress-label"),
            hg.SPAN(_class="bx--progress-line"),
        ]

        if optional:
            elements.insert(2, hg.P(_("Optional"), _class="bx--progress-optional"))

        if tooltip is not None:
            tooltipid = hg.html_id(tooltip, "tooltip-label")
            elements[1].attributes["aria-describedby"] = tooltipid
            elements.insert(
                2,
                hg.DIV(
                    hg.SPAN(_class="bx--tooltip__caret"),
                    hg.P(tooltip, _class="bx--tooltip__text"),
                    id=tooltipid,
                    role="tooltip",
                    data_floating_menu_direction="bottom",
                    _class="bx--tooltip",
                    data_avoid_focus_on_open=True,
                ),
            )

        super().__init__(*elements, **kwargs)


class ProgressIndicator(hg.UL):
    def __init__(self, steps, vertical=False, **kwargs):
        """steps: lazy object or iterator of tuples in the form (step_name, step_status)"""
        kwargs["data_progress"] = True
        kwargs["data_progress_current"] = True
        kwargs["_class"] = (
            kwargs.get("_class", "")
            + " bx--progress"
            + (" bx--progress--vertical" if vertical else "")
        )
        self.steps = steps
        super().__init__(**kwargs)

    def render(self, context):
        steps = hg.resolve_lazy(self.steps, context)
        self.extend((ProgressStep(label, status) for label, status in steps))
        return super().render(context)



import collections
import uuid

import htmlgenerator as hg
from django import forms
from django.contrib import messages
from django.contrib.messages.storage.base import Message as DjangoMessage
from django.core.paginator import Paginator
from django.utils.translation import gettext_lazy as _

from bread.layout.components import (
    button,
    content_switcher,
    datatable,
    grid,
    icon,
    loading,
    modal,
    notification,
    overflow_menu,
    pagination,
    progress_indicator,
    search,
    tabs,
    tag,
    tile,
    toggle,
    tooltip,
)
from bread.layout.components.forms import Form, FormField
from bread.utils import Link

LOREMS = (
    (
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore "
        "magna aliqua. Est lorem ipsum dolor sit amet consectetur adipiscing. Sit amet mattis vulputate enim nulla "
        "aliquet porttitor lacus luctus. Elit ullamcorper dignissim cras tincidunt. Sed risus pretium quam vulputate "
        "dignissim. Quisque sagittis purus sit amet volutpat consequat mauris nunc congue. In eu mi bibendum neque "
        "egestas. Nibh tortor id aliquet lectus proin. Egestas purus viverra accumsan in nisl. Egestas maecenas "
        "pharetra convallis posuere."
    ),
    (
        "Tincidunt augue interdum velit euismod in pellentesque massa. Et malesuada fames ac turpis egestas integer. "
        "Scelerisque varius morbi enim nunc faucibus a pellentesque. Tincidunt vitae semper quis lectus. Nisi "
        "scelerisque eu ultrices vitae auctor eu augue. Fermentum posuere urna nec tincidunt praesent. Odio tempor "
        "orci dapibus ultrices in. Risus commodo viverra maecenas accumsan lacus vel facilisis volutpat est. Lectus "
        "magna fringilla urna porttitor. Adipiscing vitae proin sagittis nisl rhoncus mattis. Consequat id porta nibh "
        "venenatis cras. Urna id volutpat lacus laoreet non curabitur gravida arcu ac."
    ),
    (
        "At ultrices mi tempus imperdiet nulla malesuada pellentesque elit eget. Imperdiet massa tincidunt nunc "
        "pulvinar sapien et ligula ullamcorper malesuada. Mollis nunc sed id semper risus. Erat nam at lectus urna "
        "duis convallis convallis tellus id. Urna condimentum mattis pellentesque id nibh tortor id aliquet. Diam ut "
        "venenatis tellus in metus vulputate eu scelerisque felis. Cursus vitae congue mauris rhoncus. Quis viverra "
        "nibh cras pulvinar mattis nunc sed blandit libero. Imperdiet proin fermentum leo vel. Tincidunt augue "
        "interdum velit euismod in pellentesque massa placerat duis. Pellentesque eu tincidunt tortor aliquam nulla "
        "facilisi."
    ),
    (
        "Interdum varius sit amet mattis vulputate enim. Dolor sit amet consectetur adipiscing elit ut aliquam purus. "
        "Nulla facilisi etiam dignissim diam quis enim. Non odio euismod lacinia at quis risus sed vulputate. Felis "
        "donec et odio pellentesque diam. Sit amet volutpat consequat mauris nunc congue nisi vitae. Elementum tempus "
        "egestas sed sed risus pretium. Lacus sed turpis tincidunt id aliquet risus feugiat. Sed id semper risus in "
        "hendrerit gravida. Habitasse platea dictumst quisque sagittis. In iaculis nunc sed augue. Semper auctor neque "
        "vitae tempus quam pellentesque. Eu lobortis elementum nibh tellus. Amet justo donec enim diam vulputate. Sit "
        "amet commodo nulla facilisi. Laoreet id donec ultrices tincidunt arcu non sodales. Integer eget aliquet nibh "
        "praesent. Feugiat vivamus at augue eget arcu dictum varius duis.",
    ),
    (
        "Enim nunc faucibus a pellentesque sit amet porttitor eget dolor. Adipiscing elit pellentesque habitant morbi "
        "tristique senectus. Sed egestas egestas fringilla phasellus faucibus scelerisque eleifend donec. Imperdiet "
        "sed euismod nisi porta lorem mollis aliquam ut porttitor. Nunc vel risus commodo viverra maecenas accumsan. "
        "Gravida dictum fusce ut placerat orci. Ac turpis egestas integer eget aliquet nibh. Aliquet risus feugiat in "
        "ante metus dictum at tempor. Viverra vitae congue eu consequat ac. Quisque sagittis purus sit amet volutpat "
        "consequat mauris. Nunc sed blandit libero volutpat sed. Vel pretium lectus quam id leo in. Nisi vitae suscipit "
        "tellus mauris a diam maecenas sed. Enim sit amet venenatis urna cursus eget nunc. Odio aenean sed adipiscing "
        "diam donec adipiscing tristique. Vivamus at augue eget arcu dictum varius. Feugiat in ante metus dictum at "
        "tempor commodo ullamcorper a.",
    ),
)


def table_of_contents(menu_list: list, show_header=True):
    """Generate a table of content to browse within the webpage."""
    ret = hg.UL(
        _class=f"bx--list--unordered {'' if show_header else 'bx--list--nested'}",
    )

    latest_list = hg.LI(_class="bx--list__item")
    for item in menu_list:
        if isinstance(item, list):
            latest_list.append(table_of_contents(item, False))
        else:
            anchor, title = item
            if len(latest_list) > 0:
                ret.append(latest_list)
                latest_list = hg.LI(_class="bx--list__item")
            latest_list.append(hg.A(title, href=f"#{anchor}"))

    if len(ret) < len(menu_list):
        ret.append(latest_list)

    if show_header:
        return grid.Row(
            hg.STYLE(
                hg.mark_safe(
                    """
                    .componentpreview-toc {
                        padding: 2rem;
                        margin-bottom: 2rem;
                    }
                    .componentpreview-toc>ul {
                        margin: 1.75rem 1.5rem;
                    }
                    .componentpreview-toc>ul>::before {
                        display: none;
                    }
                    """
                )
            ),
            grid.Col(
                tile.Tile(
                    hg.H4(_("Table of Contents")),
                    ret,
                    _class="componentpreview-toc",
                ),
                breakpoint="lg",
                width=8,
            ),
            grid.Col(breakpoint="lg", width=8),
        )

    return ret


def table_of_contents_from_cls(*classes):
    modules = collections.defaultdict(list)
    for cls in classes:
        module_name = cls.__module__.split(".")[-1]
        modules[cls.__module__].append(
            (f"{module_name}-{cls.__name__.lower()}", cls.__name__)
        )

    ret = []
    for module in modules:
        name = module.split(".")[-1]
        ret.extend(((name, module), modules[module]))

    return table_of_contents(ret)


def section_header(header, anchor=None):
    return hg.BaseElement(
        hg.A(name=anchor or header.lower()),
        hg.H1(header, style="margin-top: 6rem;"),
        hg.HR(),
    )


def section(cls, *content):
    module_name = cls.__module__.split(".")[-1]

    return hg.BaseElement(
        hg.A(name=f"{module_name}-{cls.__name__.lower()}"),
        hg.H6(cls.__module__, style="margin-top: 2rem;"),
        hg.H4(cls.__name__, style="margin-bottom: 1.5rem;"),
        hg.PRE(hg.CODE(cls.__doc__), style="margin-bottom: 1rem;"),
        *content,
    )


def layout(request):
    return hg.BaseElement(
        table_of_contents_from_cls(
            grid.Grid,
            grid.Row,
            grid.Col,
            tabs.Tabs,
            modal.Modal,
            tile.Tile,
            tile.ExpandableTile,
        ),
        _grid_py(),
        _tabs_py(),
        _modal_py(),
        _tile_py(),
    )


def informational(request):
    return hg.BaseElement(
        table_of_contents_from_cls(
            icon.Icon,
            tag.Tag,
            loading.Loading,
            progress_indicator.ProgressIndicator,
            progress_indicator.ProgressStep,
            tooltip.DefinitionTooltip,
            tooltip.IconTooltip,
            tooltip.InteractiveTooltip,
            notification.InlineNotification,
            notification.ToastNotification,
        ),
        _icon_py(),
        _tag_py(),
        _loading_py(),
        _progress_indicator_py(),
        _tooltip_py(),
        _notification_py(request),
    )


def interactive(request):
    return hg.BaseElement(
        table_of_contents_from_cls(
            button.Button,
            content_switcher.ContentSwitcher,
            toggle.Toggle,
            overflow_menu.OverflowMenu,
            pagination.Pagination,
            search.Search,
        ),
        _button_py(),
        _content_switcher_py(),
        _toggle_py(),
        _overflow_menu_py(),
        _pagination_py(),
        _searchbar_py(),
    )


def datatable_layout(request):
    return hg.BaseElement(
        table_of_contents(
            [
                ("datatable", "bread.layout.components.datatable"),
                [
                    ("datatable-datatable", "DataTable"),
                    [
                        ("datatable-datatable-with_toolbar", "with_toolbar method"),
                    ],
                ],
            ]
        ),
        _datatable_py(),
    )


def _grid_py():
    def grid_col_preview(contents, is_short=False, breakpoint="lg", width=None):
        return grid.Col(
            contents,
            style=(
                "align-items: center; "
                "background: #edf5ff; "
                "color: #000000; "
                "display: flex; "
                "justify-content: center; "
                "outline: 1px solid #a6c8ff; "
                "overflow: hidden; "
                "text-align: center; "
            )
            + ("height: 2.5rem; " if is_short else "height: 100px; "),
            breakpoint=breakpoint,
            width=width,
        )

    def grid_content_area_preview(content):
        return hg.DIV(
            content,
            style=(
                "display: flex;"
                "justify-content: center;"
                "align-items: center;"
                "background-color: #9ef0f0;"
                "height: 100px; "
                "text-align: center;"
                "width: 100%; "
            ),
        )

    def row_mode_preview(gridmode):
        return grid.Row(
            hg.Iterator(
                range(4),
                "colindex",
                grid_col_preview(grid_content_area_preview("area")),
            ),
            gridmode=gridmode,
        )

    grid_gutter_preview = hg.BaseElement(
        grid.Row(
            hg.Iterator(
                range(1, 5),
                "colindex",
                grid_col_preview(hg.F(lambda c: str(c["colindex"]) + " of 4")),
            )
        ),
        grid.Row(
            hg.Iterator(
                range(1, 3),
                "colindex",
                grid_col_preview(hg.F(lambda c: str(c["colindex"]) + " of 2")),
            )
        ),
        grid.Row(grid_col_preview("1 of 1")),
    )

    grid_mode_preview = hg.BaseElement(
        grid.Row(
            hg.Iterator(
                range(4),
                "colindex",
                grid_col_preview(grid_content_area_preview("area")),
            )
        ),
        grid.Row(
            hg.Iterator(
                range(2),
                "colindex",
                grid_col_preview(grid_content_area_preview("area")),
            )
        ),
        grid.Row(grid_col_preview(grid_content_area_preview("area"))),
    )

    breakpoints = [
        {
            "name": "sm",
            "size": 320,
            "width": 4,
        },
        {
            "name": "md",
            "size": 672,
            "width": 8,
        },
        {
            "name": "lg",
            "size": 1056,
            "width": 16,
        },
        {
            "name": "xlg",
            "size": 1312,
            "width": 16,
        },
        {
            "name": "max",
            "size": 1584,
            "width": 16,
        },
    ]

    return hg.BaseElement(
        section_header("Grid"),
        section(
            grid.Grid,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("gutter=True (default)"),
                        grid.Grid(
                            grid_gutter_preview,
                            gutter=True,
                        ),
                        style="margin-bottom: 1rem;",
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("gutter=False"),
                        grid.Grid(
                            grid_gutter_preview,
                            gutter=False,
                        ),
                        style="margin-bottom: 1rem;",
                    ),
                ),
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('gridmode="full-width" (default)'),
                        grid.Grid(
                            grid_mode_preview,
                            gridmode="full-width",
                        ),
                        style="margin-bottom: 1rem;",
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('gridmode="condensed"'),
                        grid.Grid(
                            grid_mode_preview,
                            gridmode="condensed",
                        ),
                        style="margin-bottom: 1rem;",
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('gridmode="narrow"'),
                        grid.Grid(
                            grid_mode_preview,
                            gridmode="narrow",
                        ),
                        style="margin-bottom: 1rem;",
                    ),
                ),
            ),
        ),
        section(
            grid.Row,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("gridmode=None (default)"),
                        hg.H5('* the same as "full-width"'),
                        row_mode_preview(None),
                        style="margin-bottom: 1rem;",
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('gridmode="condensed"'),
                        row_mode_preview("condensed"),
                        style="margin-bottom: 1rem;",
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('gridmode="narrow"'),
                        row_mode_preview("narrow"),
                        style="margin-bottom: 1rem;",
                    ),
                ),
            ),
        ),
        section(
            grid.Col,
            grid.Row(
                grid.Col(
                    hg.H4("Flexible width"),
                    tile.Tile(
                        hg.H4("width=None (default)"),
                        hg.P(
                            (
                                "Widths are divided equally within the same row regardless of the number of columns "
                                "inside. If the viewport is too narrow, rightmost columns may fall to the new row."
                            ),
                            style="margin-bottom: 2rem;",
                        ),
                        grid.Grid(
                            hg.Iterator(
                                range(8, 3, -1),
                                "rowindex",
                                grid.Row(
                                    hg.F(
                                        lambda cout: hg.Iterator(
                                            range(1, cout["rowindex"] + 1),
                                            "colindex",
                                            grid_col_preview(
                                                hg.F(
                                                    lambda c: "%d of %d"
                                                    % (c["colindex"], c["rowindex"])
                                                ),
                                                is_short=True,
                                            ),
                                        )
                                    ),
                                ),
                            ),
                        ),
                        style="margin-bottom: 1rem;",
                    ),
                ),
            ),
            grid.Row(
                grid.Col(
                    hg.H4("Fixed width"),
                    hg.P(
                        "Each column is assigned their minimum possible width. If widths are getting less than the ",
                        "minimum width, the rightmost columns will fall as a new row instead. You can learn more ",
                        "about possible widths (number of columns) ",
                        hg.A(
                            "here",
                            href="https://www.carbondesignsystem.com/guidelines/2x-grid/overview/#breakpoints",
                            rel="noreferrer noopener",
                            target="_blank",
                        ),
                        ". You can also resize the window (or, alternatively, zoom in-out) to see how columns will ",
                        "rearrange.",
                        style="margin-bottom: 2rem;",
                    ),
                ),
            ),
            hg.Iterator(
                breakpoints,
                "bp",
                grid.Row(
                    grid.Col(
                        tile.Tile(
                            hg.H4(hg.F(lambda c: f'breakpoint="{c["bp"]["name"]}"')),
                            hg.P(
                                hg.F(
                                    lambda c: (
                                        f"The minimum viewport width is {c['bp']['size']} px. "
                                        f"The maximum total width for \"{c['bp']['name']}\" is {c['bp']['width']}. "
                                        "Shorter widths of viewport will result in rightmost columns would "
                                        " fall as a new row instead."
                                    )
                                ),
                                style="margin-bottom: 2rem;",
                            ),
                            grid.Grid(
                                hg.F(
                                    lambda cout: hg.Iterator(
                                        range(cout["bp"]["width"], 0, -1),
                                        "rowindex",
                                        grid.Row(
                                            hg.F(
                                                lambda c: hg.BaseElement(
                                                    hg.Iterator(
                                                        range(1, c["rowindex"]),
                                                        "colindex",
                                                        grid_col_preview(
                                                            "width=1",
                                                            is_short=True,
                                                        ),
                                                    ),
                                                    grid_col_preview(
                                                        "width=%d"
                                                        % (
                                                            c["bp"]["width"]
                                                            - c["rowindex"]
                                                            + 1
                                                        ),
                                                        is_short=True,
                                                        breakpoint=c["bp"]["name"],
                                                        width=c["bp"]["width"]
                                                        - c["rowindex"]
                                                        + 1,
                                                    ),
                                                ),
                                            ),
                                        ),
                                    )
                                ),
                                style="margin-bottom: 1rem;",
                            ),
                        ),
                    ),
                    style="margin-bottom: 1rem;",
                ),
            ),
        ),
    )


def _tabs_py():
    def sample_tabs(*tabnames):
        return tuple(
            tabs.Tab(
                f"Tab {tabname}",
                hg.BaseElement(
                    hg.P(
                        f"This is the content within tab {tabname}.",
                        style="margin-bottom: 0.5rem;",
                    ),
                    hg.P(dummytxt),
                ),
            )
            for tabname, dummytxt in zip(tabnames, LOREMS)
        )

    return hg.BaseElement(
        section_header("Tabs"),
        section(
            tabs.Tabs,
            grid.Row(
                # Use different tab name to avoid the tab switching confusion.
                grid.Col(
                    tile.Tile(
                        hg.H4("container=False (default)"),
                        tabs.Tabs(*sample_tabs("1", "2", "3"), container=False),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("container=True"),
                        tabs.Tabs(*sample_tabs("A", "B", "C"), container=True),
                    ),
                ),
            ),
            grid.Row(
                grid.Col(),
            ),
        ),
    )


def _modal_py():
    def gen_sample_modal(size):
        return modal.Modal(
            "Modal Heading",
            hg.Iterator(
                LOREMS,
                "dummy_paragraph",
                hg.P(
                    hg.F(lambda c: c["dummy_paragraph"]),
                    style="margin-bottom: 0.5rem;",
                ),
            ),
            label='Sample Label with size="%s"' % size,
            size=size,
            buttons=(
                button.Button("OK"),
                button.Button("Action 1", buttontype="secondary"),
                hg.If(size != "xs", button.Button("Action 2", buttontype="secondary")),
                button.Button("Cancel", buttontype="danger", data_modal_close=True),
            ),
        )

    sizes = "xs", "sm", "md", "lg"
    sample_modal = tuple((size, gen_sample_modal(size)) for size in sizes)

    return hg.BaseElement(
        section_header("Modal"),
        section(
            modal.Modal,
            hg.Iterator(
                sample_modal,
                "sample_modal",
                hg.F(
                    lambda c: button.Button(
                        'Open a modal of size="%s"' % c["sample_modal"][0],
                        style="margin-right: 1rem;",
                        **c["sample_modal"][1].openerattributes,
                    )
                ),
            ),
            hg.Iterator(
                sample_modal, "sample_modal", hg.F(lambda c: c["sample_modal"][1])
            ),
        ),
    )


def _tile_py():
    return hg.BaseElement(
        section_header("Tile"),
        grid.Row(
            grid.Col(
                section(
                    tile.Tile,
                    tile.Tile(
                        hg.P("This is a content in Tile."),
                        hg.P("Almost everything can be put here."),
                    ),
                ),
            ),
            grid.Col(
                section(
                    tile.ExpandableTile,
                    tile.ExpandableTile(
                        hg.DIV(
                            hg.P(
                                "This is the visible part of ExpandableTile. Click to view the hidden one."
                            ),
                        ),
                        hg.DIV(
                            hg.P("This is a hidden part of ExpandableTile."),
                        ),
                        {"style": "height: 100px;"},
                        {"style": "height: 100px;"},
                    ),
                ),
            ),
        ),
    )


def _icon_py():
    return hg.BaseElement(
        section_header("Icon"),
        section(
            icon.Icon,
            grid.Row(
                hg.Iterator(
                    ("information", "filter", "email"),
                    "iconname",
                    grid.Col(
                        tile.Tile(
                            hg.H4('name="', hg.C("iconname"), '"'),
                            icon.Icon(hg.C("iconname")),
                        )
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("name=[iconname]"),
                        hg.H6("* required"),
                        hg.P("display the icon corresponding to iconname"),
                    )
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("size=None (default)"),
                        hg.H6('* will be interpreted as size="32" (in pixel)'),
                        icon.Icon("information"),
                        icon.Icon("filter"),
                        icon.Icon("email"),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('size="16"'),
                        icon.Icon("information", size="16"),
                        icon.Icon("filter", size="16"),
                        icon.Icon("email", size="16"),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('size="64"'),
                        icon.Icon("information", size="64"),
                        icon.Icon("filter", size="64"),
                        icon.Icon("email", size="64"),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("size=[preferred length]"),
                        hg.P("the width and height of the icon"),
                    )
                ),
            ),
        ),
    )


def _tag_py():
    return hg.BaseElement(
        section_header("Tag"),
        section(
            tag.Tag,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("can_delete=False (default)"),
                        tag.Tag("image", tag_color="warm-gray"),
                        tag.Tag("document", tag_color="cyan"),
                        tag.Tag("important", tag_color="magenta"),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("can_delete=True"),
                        tag.Tag("image", can_delete=True, tag_color="warm-gray"),
                        tag.Tag("document", can_delete=True, tag_color="cyan"),
                        tag.Tag("important", can_delete=True, tag_color="magenta"),
                    )
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("tag_color=None (default)"),
                        hg.H6('the same as tag_color="gray"'),
                        tag.Tag("image"),
                    ),
                    breakpoint="lg",
                    width=4,
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('Possible "tag_color" value'),
                        hg.Iterator(
                            (
                                "red",
                                "magenta",
                                "purple",
                                "blue",
                                "cyan",
                                "teal",
                                "green",
                                "gray",
                                "cool-gray",
                                "warm-gray",
                            ),
                            "tag_color",
                            hg.F(
                                lambda c: tag.Tag(
                                    c["tag_color"], tag_color=c["tag_color"]
                                )
                            ),
                        ),
                    ),
                    breakpoint="md",
                    width=6,
                ),
            ),
        ),
    )


def _loading_py():
    return hg.BaseElement(
        section_header("Loading"),
        section(
            loading.Loading,
            grid.Row(
                grid.Col(tile.Tile(hg.H4("small=False (default)"), loading.Loading())),
                grid.Col(tile.Tile(hg.H4("small=True"), loading.Loading(small=True))),
            ),
        ),
    )


def _progress_indicator_py():
    progressstep_demo_optional = progress_indicator.ProgressIndicator(
        (("Another step", "incomplete"),)
    )
    progressstep_demo_optional.extend(
        (
            progress_indicator.ProgressStep("=False", "complete", optional=False),
            progress_indicator.ProgressStep("=True", "current", optional=True),
        )
    )
    progressstep_demo_disabled = progress_indicator.ProgressIndicator(
        (("Another step", "current"),)
    )
    progressstep_demo_disabled.extend(
        (
            progress_indicator.ProgressStep("=False", "complete", disabled=False),
            progress_indicator.ProgressStep("=True", "incomplete", disabled=True),
        )
    )

    return hg.BaseElement(
        section_header("Progress Indicator", "progress_indicator"),
        section(
            progress_indicator.ProgressIndicator,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('Possible "status" value'),
                        hg.H6("* required"),
                        progress_indicator.ProgressIndicator(
                            (status, status)
                            for status in (
                                "warning",
                                "current",
                                "complete",
                                "incomplete",
                            )
                        ),
                    )
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("vertical=False (default)"),
                        progress_indicator.ProgressIndicator(
                            (
                                ("Step 1", "complete"),
                                ("Step 2", "current"),
                                ("Step 3", "incomplete"),
                                ("Step 4", "incomplete"),
                            )
                        ),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("vertical=True"),
                        progress_indicator.ProgressIndicator(
                            (
                                ("Step 1", "complete"),
                                ("Step 2", "current"),
                                ("Step 3", "incomplete"),
                                ("Step 4", "incomplete"),
                            ),
                            vertical=True,
                        ),
                    )
                ),
            ),
        ),
        section(
            progress_indicator.ProgressStep,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('Possible "disabled" value'),
                        hg.H6("* default is False"),
                        progressstep_demo_disabled,
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('Possible "optional" value'),
                        hg.H6("* default is False"),
                        progressstep_demo_optional,
                    )
                ),
                style="margin-bottom: 2rem;",
            ),
        ),
    )


def _tooltip_py():
    return hg.BaseElement(
        section_header("Tooltip"),
        grid.Row(
            grid.Col(
                section(
                    tooltip.DefinitionTooltip,
                    tile.Tile(
                        hg.DIV(
                            tooltip.DefinitionTooltip(
                                "Definition tooltip (left aligned)",
                                "Brief definition of the dotted, underlined word above.",
                                align="left",
                            )
                        ),
                        hg.DIV(
                            tooltip.DefinitionTooltip(
                                "Definition tooltip (center aligned)",
                                "Brief definition of the dotted, underlined word above.",
                                align="center",
                            )
                        ),
                        hg.DIV(
                            tooltip.DefinitionTooltip(
                                "Definition tooltip (right aligned)",
                                "Brief definition of the dotted, underlined word above.",
                                align="right",
                            )
                        ),
                    ),
                ),
                section(
                    tooltip.IconTooltip,
                    tile.Tile(
                        tooltip.IconTooltip(
                            "Help",
                        ),
                        tooltip.IconTooltip(
                            "Filter",
                            icon=icon.Icon("filter"),
                        ),
                        tooltip.IconTooltip(
                            "Email",
                            icon="email",
                        ),
                    ),
                ),
                section(
                    tooltip.InteractiveTooltip,
                    tile.Tile(
                        tooltip.InteractiveTooltip(
                            label="Tooltip label",
                            body=(
                                _(
                                    "This is some tooltip text. This box shows the maximum amount of text that should "
                                    "appear inside. If more room is needed please use a modal instead."
                                )
                            ),
                            heading="Heading within a Tooltip",
                            button=(button.Button("Button")),
                            link=Link(href="#", label="link"),
                        ),
                    ),
                ),
            ),
        ),
    )


class ToastNotificationPreview(notification.ToastNotification):
    """
    The class that override some attributes of ToastNotification for the purpose of component preview.
    It can be used just like ToastNotification, but some functionalities are disabled.
    """

    def __init__(
        self,
        message,
        details,
        kind="info",
        lowcontrast=False,
        hideclosebutton=False,
        hidetimestamp=False,
        autoremove=4.0,
        **attributes,
    ):
        attributes["style"] = hg.BaseElement(
            ";opacity: 1 !important;animation: unset !important;"
        )
        attributes["onload"] = "//"
        super().__init__(
            message,
            details,
            kind,
            lowcontrast,
            hideclosebutton,
            hidetimestamp,
            autoremove,
            **attributes,
        )


def _notification_py(request):
    class TriggerToastMessageForm(forms.Form):
        displaymessage = forms.BooleanField(
            widget=forms.HiddenInput,
            initial=True,
        )

    form: TriggerToastMessageForm
    received_post = False

    if request.method == "POST":
        form = TriggerToastMessageForm(request.POST)
        if form.is_valid() and "displaymessage" in form.cleaned_data:
            received_post = True
            messages.info(
                request, _("This is a server message in the ToastNotification.")
            )

    if not received_post:
        form = TriggerToastMessageForm()

    # for ToastNotification
    message_obj = (DjangoMessage("20", "This is a detail."),)
    notification_kinds = (
        "error",
        "info",
        "info-square",
        "success",
        "warning",
        "warning-alt",
    )
    toast_kinds = (
        hg.Iterator(
            message_obj,
            "message",
            ToastNotificationPreview(
                "This is a message.", hg.C("message.message"), kind=kind
            ),
        )
        for kind in notification_kinds
    )
    toast_demo_kind = hg.BaseElement(
        *(
            grid.Col(
                tile.Tile(
                    hg.H4(f'kind="{kind}"'),
                    toast,
                ),
                style="margin-bottom: 2rem;",
            )
            for kind, toast in zip(notification_kinds, toast_kinds)
        )
    )

    return hg.BaseElement(
        section_header("Notification"),
        section(
            notification.InlineNotification,
            notification.InlineNotification(
                "This is a message.",
                "This is a detail.",
            ),
            notification.InlineNotification(
                "This is an InlineNotification with an action.",
                "Click the link to learn more.",
                action=(
                    "Learn more",
                    "window.alert('This is a possible onclick event for InlineNotification.');",
                ),
            ),
            grid.Row(
                hg.Iterator(
                    notification_kinds,
                    "kind",
                    grid.Col(
                        tile.Tile(
                            hg.H4('kind="', hg.C("kind"), '"'),
                            notification.InlineNotification(
                                "This is a message.",
                                "This is a detail.",
                                kind=hg.C("kind"),
                            ),
                            breakpoint="md",
                            width=4,
                        ),
                        style="margin-bottom: 2rem;",
                    ),
                ),
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("lowcontrast=False (default)"),
                        notification.InlineNotification(
                            "This is a message.",
                            "This is a detail.",
                            lowcontrast=False,
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("lowcontrast=True"),
                        notification.InlineNotification(
                            "This is a message.",
                            "This is a detail.",
                            lowcontrast=True,
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("hideclosebutton=False (default)"),
                        notification.InlineNotification(
                            "This is a message.",
                            "This is a detail.",
                            hideclosebutton=False,
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("hideclosebutton=True"),
                        notification.InlineNotification(
                            "This is a message.",
                            "This is a detail.",
                            hideclosebutton=True,
                        ),
                    ),
                ),
            ),
        ),
        section(
            notification.ToastNotification,
            hg.Iterator(
                message_obj,
                "message",
                ToastNotificationPreview(
                    "This is a message.",
                    hg.C("message.message"),
                ),
            ),
            Form(
                form,
                FormField("displaymessage"),
                button.Button(
                    _("Click to see ToastNotification in action"), type="submit"
                ),
            ),
            hg.DIV(
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                toast_demo_kind,
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("lowcontrast=False (default)"),
                        hg.Iterator(
                            message_obj,
                            "message",
                            ToastNotificationPreview(
                                "This is a message.",
                                hg.C("message.message"),
                                lowcontrast=False,
                            ),
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("lowcontrast=True"),
                        hg.Iterator(
                            message_obj,
                            "message",
                            ToastNotificationPreview(
                                "This is a message.",
                                hg.C("message.message"),
                                lowcontrast=True,
                            ),
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("hideclosebutton=False (default)"),
                        hg.Iterator(
                            message_obj,
                            "message",
                            ToastNotificationPreview(
                                "This is a message.",
                                hg.C("message.message"),
                                hideclosebutton=False,
                            ),
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("hideclosebutton=True"),
                        hg.Iterator(
                            message_obj,
                            "message",
                            ToastNotificationPreview(
                                "This is a message.",
                                hg.C("message.message"),
                                hideclosebutton=True,
                            ),
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("hidetimestamp=False (default)"),
                        hg.Iterator(
                            message_obj,
                            "message",
                            ToastNotificationPreview(
                                "This is a message.",
                                hg.C("message.message"),
                                hidetimestamp=False,
                            ),
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("hidetimestamp=True"),
                        hg.Iterator(
                            message_obj,
                            "message",
                            ToastNotificationPreview(
                                "This is a message.",
                                hg.C("message.message"),
                                hidetimestamp=True,
                            ),
                        ),
                    ),
                ),
            ),
        ),
    )


def _button_py():
    button_types = (
        "primary",
        "secondary",
        "tertiary",
        "danger",
        "ghost",
    )

    return hg.BaseElement(
        section_header("Button"),
        section(
            button.Button,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('buttontype="primary" (default)'),
                        button.Button("Button"),
                    ),
                    breakpoint="md",
                    width="3",
                    style="margin-bottom: 2rem;",
                ),
                hg.Iterator(
                    button_types[1 : len(button_types) - 1],
                    "button_type",
                    grid.Col(
                        tile.Tile(
                            hg.H4('buttontype="', hg.C("button_type"), '"'),
                            hg.F(
                                lambda c: button.Button(
                                    "Button", buttontype=c["button_type"]
                                )
                            ),
                            style="margin-bottom: 2rem;",
                        ),
                        breakpoint="md",
                        width="2",
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('buttontype="ghost"'),
                        hg.H6("* transparent background"),
                        button.Button("Button", buttontype="ghost"),
                    ),
                    breakpoint="md",
                    width="2",
                    style="margin-bottom: 2rem;",
                ),
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("icon=None (default)"),
                        button.Button("Button", icon=None),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("icon=[icon name or Icon instance]"),
                        hg.H6('(for example, this one is icon="information")'),
                        button.Button("Button", icon="information"),
                    )
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("notext=False (default)"),
                        button.Button(
                            "Button",
                            icon="information",
                            notext=False,
                        ),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("notext=True"),
                        button.Button(
                            "Button",
                            icon="information",
                            notext=True,
                        ),
                    )
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("small=False (default)"),
                        button.Button("Button", small=False),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("small=True"),
                        button.Button("Button", small=True),
                    )
                ),
            ),
        ),
    )


def _content_switcher_py():
    return hg.BaseElement(
        section_header("Content Switcher", "content_switcher"),
        section(
            content_switcher.ContentSwitcher,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        content_switcher.ContentSwitcher(
                            ("Mode 1", {}),
                            ("Mode 2", {}),
                            ("Mode 3", {}),
                        )
                    ),
                    style="margin-bottom: 2rem;",
                )
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("selected=0 (default)"),
                        content_switcher.ContentSwitcher(
                            ("Mode 1", {}),
                            ("Mode 2", {}),
                            ("Mode 3", {}),
                        ),
                    ),
                    style="margin-bottom: 2rem;",
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("selected=1"),
                        content_switcher.ContentSwitcher(
                            ("Mode 1", {}), ("Mode 2", {}), ("Mode 3", {}), selected=1
                        ),
                    ),
                    style="margin-bottom: 2rem;",
                ),
                grid.Col(
                    hg.H4("selected=[index of label]"),
                    hg.P("Make the corresponding button selected by default."),
                    style="margin-bottom: 2rem;",
                ),
            ),
        ),
    )


def _toggle_py():
    return hg.BaseElement(
        section_header("Toggle"),
        section(
            toggle.Toggle,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('offlabel=_("Off") (default)'),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("offlabel=[label when toggle is off]"),
                        toggle.Toggle(
                            "Label",
                            offlabel="I'm really off!",
                            widgetattributes={"id": uuid.uuid4()},
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('onlabel=_("On") (default)'),
                        toggle.Toggle(
                            "Label",
                            checked=True,
                            widgetattributes={"id": uuid.uuid4(), "checked": True},
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4(
                            "onlabel=[label when toggle is on]",
                        ),
                        toggle.Toggle(
                            "Label",
                            onlabel="I'm really on!",
                            checked=True,
                            widgetattributes={"id": uuid.uuid4(), "checked": True},
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("help_text=None (default)"),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("help_text=[text to describe user]"),
                        toggle.Toggle(
                            "Label",
                            help_text="This is the help text.",
                            widgetattributes={"id": uuid.uuid4()},
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("errors=None (default)"),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("errors=[list of requirements]"),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                            errors=[
                                "All required fields must be filled.",
                            ],
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("disabled=None (default)"),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("disabled=True"),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                            disabled=True,
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("required=None (default)"),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                        ),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("required=True"),
                        toggle.Toggle(
                            "Label",
                            widgetattributes={"id": uuid.uuid4()},
                            required=True,
                        ),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
        ),
    )


def _overflow_menu_py():
    default_links = (
        Link(href="#overflow_menu-overflowmenu", label="Link 1", iconname=None),
        Link(
            href="#overflow_menu-overflowmenu", label="Link 2", iconname="information"
        ),
        Link(href="#overflow_menu-overflowmenu", label="Link 3"),
    )
    return hg.BaseElement(
        section_header("Overflow Menu", "overflow_menu"),
        section(
            overflow_menu.OverflowMenu,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.P(
                            "Click the three dots below to see overflow menu in action."
                        ),
                        overflow_menu.OverflowMenu(links=default_links),
                    ),
                    breakpoint="md",
                    width=2,
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("menuname=None (default)"),
                        overflow_menu.OverflowMenu(links=default_links),
                    ),
                    breakpoint="md",
                    width=3,
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("menuname=[name for menu]"),
                        overflow_menu.OverflowMenu(
                            links=default_links, menuname="This is a menu name."
                        ),
                    ),
                    breakpoint="md",
                    width=3,
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('direction="bottom" (default)'),
                        overflow_menu.OverflowMenu(links=default_links),
                    ),
                    breakpoint="md",
                    width=2,
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('direction="right"'),
                        overflow_menu.OverflowMenu(
                            links=default_links, direction="right"
                        ),
                    ),
                    breakpoint="md",
                    width=2,
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('direction="top"'),
                        overflow_menu.OverflowMenu(
                            links=default_links, direction="top"
                        ),
                    ),
                    breakpoint="md",
                    width=2,
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('direction="left"'),
                        overflow_menu.OverflowMenu(
                            links=default_links, direction="left"
                        ),
                    ),
                    breakpoint="md",
                    width=2,
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("flip=False (default)"),
                        overflow_menu.OverflowMenu(default_links, flip=False),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("flip=True"),
                        overflow_menu.OverflowMenu(default_links, flip=True),
                    )
                ),
                style="margin-bottom: 2rem;",
            ),
        ),
    )


def _pagination_py():
    return hg.BaseElement(
        section_header("Pagination"),
        section(
            pagination.Pagination,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        pagination.Pagination(
                            Paginator(range(1, 6), 3),
                            (3, 4),
                        ),
                    ),
                ),
            ),
        ),
    )


def _searchbar_py():
    sizes = "sm", "lg", "xl"
    return hg.BaseElement(
        section_header("Search", "search"),
        section(
            search.Search,
            grid.Row(
                hg.Iterator(
                    sizes[:2],
                    "size",
                    grid.Col(
                        tile.Tile(
                            hg.H4('size="', hg.C("size"), '"'),
                            hg.F(lambda c: search.Search(size=c["size"])),
                        ),
                        breakpoint="md",
                        width=4,
                        style="margin-bottom: 2rem;",
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4('size="xl" (default)'),
                        hg.F(lambda c: search.Search(size="xl")),
                    ),
                    breakpoint="md",
                    width=4,
                    style="margin-bottom: 2rem;",
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("placeholder=None (default)"),
                        search.Search(),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("placeholder=[custom placeholder]"),
                        search.Search(placeholder="I'm the custom placeholder."),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("disabled=False (default)"),
                        search.Search(),
                    ),
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("disabled=True"),
                        search.Search(disabled=True),
                    ),
                ),
                style="margin-bottom: 2rem;",
            ),
        ),
    )


def _datatable_py():
    sample_headers = ["Country", "Continent", "Population"]
    sample_rows = [
        ["Switzerland", "Europe", 8_500_000],
        ["Germany", "Europe", 83_000_000],
        ["Thailand", "Asia", 70_000_000],
    ]

    sample_columns = [
        datatable.DataTableColumn(header=header, cell=hg.DIV(hg.C(f"row.{header}")))
        for header in sample_headers
    ]

    sample_row_iterator = [
        {header: content for header, content in zip(sample_headers, row)}
        for row in sample_rows
    ]

    return hg.BaseElement(
        section_header("Data Table", "datatable"),
        section(
            datatable.DataTable,
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4('spacing="default" (default)'),
                        datatable.DataTable(
                            sample_columns,
                            sample_row_iterator,
                        ),
                    ),
                    breakpoint="md",
                    width=4,
                    style="margin-bottom: 2rem;",
                ),
                hg.Iterator(
                    datatable.DataTable.SPACINGS[1:],
                    "spacing",
                    hg.F(
                        lambda c: hg.BaseElement(
                            grid.Col(
                                tile.Tile(
                                    hg.H4('spacing="', c["spacing"], '"'),
                                    datatable.DataTable(
                                        sample_columns,
                                        sample_row_iterator,
                                        spacing=c["spacing"],
                                    ),
                                ),
                                breakpoint="md",
                                width=4,
                                style="margin-bottom: 2rem;",
                            ),
                        ),
                    ),
                ),
            ),
            grid.Row(
                grid.Col(
                    tile.Tile(
                        hg.H4("zebra=False (default)"),
                        datatable.DataTable(
                            sample_columns,
                            sample_row_iterator,
                            zebra=False,
                        ),
                    )
                ),
                grid.Col(
                    tile.Tile(
                        hg.H4("zebra=True"),
                        datatable.DataTable(
                            sample_columns,
                            sample_row_iterator,
                            zebra=True,
                        ),
                    )
                ),
            ),
        ),
        hg.BaseElement(
            hg.A(name="datatable-datatable-with_toolbar"),
            hg.H6("bread.layout.components.datatable", style="margin-top: 2rem;"),
            hg.H4("DataTable.with_toolbar(self)", style="margin-bottom: 1.5rem;"),
            hg.PRE(
                hg.CODE(datatable.DataTable.with_toolbar.__doc__),
                style="margin-bottom: 1rem;",
            ),
            datatable.DataTable(sample_columns, sample_row_iterator).with_toolbar(
                title="Data Table Title",
                helper_text="Data Table Subtitle",
                primary_button=button.Button("Primary Button"),
                bulkactions=[
                    Link(href="#", label="action 1"),
                    Link(href="#", label="action 2"),
                    Link(href="#", label="action 3"),
                ],
            ),
        ),
        section(
            datatable.DataTableColumn,
            hg.P("This is a data table column."),
        ),
    )



import htmlgenerator as hg

import bread

from ..utils import HasBreadCookieValue
from .icon import Icon


def isactive(itemaccessor):
    return hg.F(lambda c: c[itemaccessor].active(c["request"]))


class SideNav(hg.ASIDE):
    def __init__(self, menu: "bread.menu.Menu", **kwargs):
        kwargs["_class"] = hg.BaseElement(
            kwargs.get("_class", ""),
            " bx--side-nav bx--side-nav--rail",
            hg.If(
                HasBreadCookieValue("sidenav-hidden", "true"),
                "",
                " bx--side-nav--expanded",
            ),
        )
        kwargs["data_side_nav"] = True
        super().__init__(
            hg.NAV(
                hg.UL(
                    hg.Iterator(
                        hg.F(
                            lambda c: (
                                i
                                for i in sorted(
                                    hg.resolve_lazy(menu, c)._registry.values()
                                )
                                if i.has_permission(c["request"])
                            )
                        ),
                        "menugroup",
                        hg.LI(
                            hg.If(
                                hg.F(
                                    lambda c: len(c["menugroup"].items) > 1
                                    or c["menugroup"].force_show
                                ),
                                hg.BaseElement(
                                    hg.BUTTON(
                                        hg.DIV(
                                            Icon(hg.C("menugroup.iconname"), size=16),
                                            _class="bx--side-nav__icon",
                                        ),
                                        hg.SPAN(
                                            hg.C("menugroup.label"),
                                            _class="bx--side-nav__submenu-title",
                                        ),
                                        hg.DIV(
                                            Icon("chevron--down", size=16),
                                            _class="bx--side-nav__icon bx--side-nav__submenu-chevron",
                                        ),
                                        _class="bx--side-nav__submenu",
                                        type="button",
                                        aria_haspopup="true",
                                        aria_expanded=hg.If(
                                            isactive("menugroup"), "true"
                                        ),
                                    ),
                                    hg.UL(
                                        hg.Iterator(
                                            hg.F(
                                                lambda c: (
                                                    i
                                                    for i in sorted(
                                                        c["menugroup"].items
                                                    )
                                                    if i.has_permission(c["request"])
                                                )
                                            ),
                                            "menuitem",
                                            hg.LI(
                                                hg.A(
                                                    hg.SPAN(
                                                        hg.C("menuitem.link.label"),
                                                        _class="bx--side-nav__link-text",
                                                    ),
                                                    _class=hg.BaseElement(
                                                        "bx--side-nav__link",
                                                        hg.If(
                                                            isactive("menuitem"),
                                                            " bx--side-nav__link--current",
                                                        ),
                                                    ),
                                                    href=hg.C("menuitem.link.href"),
                                                ),
                                                _class=hg.BaseElement(
                                                    "bx--side-nav__menu-item",
                                                    hg.If(
                                                        isactive("menuitem"),
                                                        " bx--side-nav__menu-item--current",
                                                    ),
                                                ),
                                            ),
                                        ),
                                        _class="bx--side-nav__menu",
                                    ),
                                ),
                                hg.A(
                                    hg.DIV(
                                        Icon(
                                            hg.C("menugroup.items.0.link.iconname"),
                                            size=16,
                                        ),
                                        _class="bx--side-nav__icon",
                                    ),
                                    hg.SPAN(
                                        hg.C("menugroup.items.0.link.label"),
                                        _class="bx--side-nav__link-text",
                                    ),
                                    _class=hg.BaseElement(
                                        "bx--side-nav__link",
                                        hg.If(
                                            isactive("menugroup"),
                                            " bx--side-nav__link--current",
                                        ),
                                    ),
                                    href=hg.C("menugroup.items.0.link.href"),
                                ),
                            ),
                            _class=hg.BaseElement(
                                "bx--side-nav__item",
                                hg.If(
                                    isactive("menugroup"), " bx--side-nav__item--active"
                                ),
                            ),
                        ),
                    ),
                    _class="bx--side-nav__items",
                ),
                hg.FOOTER(
                    hg.BUTTON(
                        hg.DIV(
                            Icon(
                                "close",
                                size=20,
                                _class="bx--side-nav__icon--collapse bx--side-nav-collapse-icon",
                                aria_hidden="true",
                            ),
                            Icon(
                                "chevron--right",
                                size=20,
                                _class="bx--side-nav__icon--expand bx--side-nav-expand-icon",
                                aria_hidden="true",
                            ),
                            _class="bx--side-nav__icon",
                        ),
                        hg.SPAN(
                            "Toggle the expansion state of the navigation",
                            _class="bx--assistive-text",
                        ),
                        _class="bx--side-nav__toggle",
                        onclick="setBreadCookie('sidenav-hidden', getBreadCookie('sidenav-hidden', 'false') != 'true');",
                    ),
                    _class="bx--side-nav__footer",
                ),
                _class="bx--side-nav__navigation",
            ),
            **kwargs,
        )



import collections
import typing

import htmlgenerator as hg
from django.utils.text import slugify

from ..utils import HasBreadCookieValue


class TabLabel(hg.LI):
    def __init__(self, label, tabid, panelid, selected):
        super().__init__(
            hg.A(
                label,
                tabindex="0",
                id=tabid,
                _class="bx--tabs__nav-link",
                href="javascript:void(0)",
                aria_controls=panelid,
                aria_selected=hg.If(selected, "true", "false"),
                role="tab",
            ),
            _class=hg.BaseElement(
                "bx--tabs__nav-item",
                hg.If(selected, " bx--tabs__nav-item--selected"),
            ),
            data_target="#" + panelid,
            aria_selected=hg.If(selected, "true", "false"),
            role="tab",
            onclick=hg.BaseElement("setBreadCookie('selected-tab', '", tabid, "')"),
        )


class TabPanel(hg.DIV):
    def __init__(self, content, panelid, tabid, selected):
        super().__init__(
            content,
            id=panelid,
            aria_labelledby=tabid,
            role="tabpanel",
            aria_hidden=hg.If(selected, "false", "true"),
            hidden=hg.If(selected, False, True),
        )


class Tab(typing.NamedTuple):
    label: typing.Any
    content: typing.Any


class Tabs(hg.DIV):
    def __init__(
        self,
        *tabs,
        container=False,
        tabpanel_attributes=None,
        labelcontainer_attributes=None,
        **attributes,
    ):
        tabpanel_attributes = collections.defaultdict(str, tabpanel_attributes or {})
        labelcontainer_attributes = collections.defaultdict(
            str, labelcontainer_attributes or {}
        )

        self.tablabels = hg.UL(
            _class="bx--tabs__nav bx--tabs__nav--hidden", style="flex-wrap: wrap;"
        )
        labelcontainer_attributes["_class"] += " bx--tabs" + (
            " bx--tabs--container" if container else ""
        )
        self.labelcontainer = hg.DIV(
            hg.DIV(
                hg.A(
                    href="javascript:void(0)",
                    _class="bx--tabs-trigger-text",
                    tabindex=-1,
                ),
                _class="bx--tabs-trigger",
                style="display: none",
                tabindex=0,
            ),
            self.tablabels,
            data_tabs=True,
            **hg.merge_html_attrs(
                {
                    "onload": "if( $('.bx--tabs__nav-item--selected', this) == null ) "
                    "(new CarbonComponents.Tab(this)).setActive($('.bx--tabs__nav-item', this))",
                },
                labelcontainer_attributes,
            ),
        )
        tabpanel_attributes["_class"] += " bx--tab-content"
        self.tabpanels = hg.DIV(**tabpanel_attributes)

        firsttab = None
        for i, (label, content) in enumerate(tabs):
            tabid = f"tab-{slugify(label)}-{i}"
            panelid = f"panel-{slugify(label)}-{i}"
            if firsttab is None:
                firsttab = tabid
            self.tablabels.append(
                TabLabel(
                    label,
                    tabid,
                    panelid,
                    HasBreadCookieValue("selected-tab", tabid, firsttab),
                )
            )
            self.tabpanels.append(
                TabPanel(
                    content,
                    panelid,
                    tabid,
                    HasBreadCookieValue("selected-tab", tabid, firsttab),
                )
            )
        super().__init__(
            self.labelcontainer,
            self.tabpanels,
        )



import base64

from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.hashers import check_password
from django.contrib.auth.models import User


class BasicHTTPAuthentication(BaseBackend):
    """
    Tries to authenticate a user via basic HTTP Authentication
    """

    def authenticate(self, request, **kwargs):
        if request is None:
            return None
        auth_header = request.headers.get("Authorization")
        if auth_header is None or not auth_header.startswith("Basic "):
            return None
        credentials_encoded = auth_header.split("Basic ", 1)[1]
        credentials = base64.b64decode(credentials_encoded).decode()
        username, password = credentials.split(":", 1)
        user = User.objects.filter(username=username).first()
        if user and check_password(password, user.password) and user.is_active:
            return user
        return None

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



import htmlgenerator as hg

from .icon import Icon


def asoverflowbutton(context):
    link = context["link"]
    return hg.A(
        hg.DIV(
            hg.If(
                link.iconname,
                Icon(link.iconname, size=16),
            ),
            link.label,
            _class="bx--overflow-menu-options__option-content",
        ),
        _class="bx--overflow-menu-options__btn",
        role="menuitem",
        title=link.label,
        href=link.href,
        **link.attributes,
    )


class OverflowMenu(hg.DIV):
    """Implements https://www.carbondesignsystem.com/components/overflow-menu/usage"""

    MENUID_TEMPLATE = "overflow-menu-%s"

    def __init__(
        self,
        links,
        menuiconname="overflow-menu--vertical",
        menuname=None,
        direction="bottom",
        flip=False,
        item_attributes={},
        **attributes,
    ):
        attributes["data-overflow-menu"] = True
        attributes["_class"] = attributes.get("_class", "") + " bx--overflow-menu"
        item_attributes["_class"] = (
            item_attributes.get("_class", "") + " bx--overflow-menu-options__option"
        )

        menuid = hg.F(
            lambda c: OverflowMenu.MENUID_TEMPLATE % hg.html_id(c.get("row", self))
        )
        triggerid = hg.F(
            lambda c: (OverflowMenu.MENUID_TEMPLATE % hg.html_id(c.get("row", self)))
            + "-trigger"
        )

        super().__init__(
            hg.BUTTON(
                Icon(menuiconname, size=16),
                _class="bx--overflow-menu__trigger"
                + (
                    " bx--tooltip__trigger bx--tooltip--a11y bx--tooltip--right bx--tooltip--align-start"
                    if menuname is not None
                    else ""
                ),
                aria_haspopup="true",
                aria_expanded="false",
                aria_controls=menuid,
                type="button",
                id=triggerid,
            ),
            hg.DIV(
                hg.UL(
                    hg.Iterator(
                        links,
                        "link",
                        hg.LI(
                            hg.F(asoverflowbutton),
                            **item_attributes,
                        ),
                    ),
                    _class="bx--overflow-menu-options__content",
                ),
                _class="bx--overflow-menu-options"
                + (" bx--overflow-menu--flip" if flip else ""),
                tabindex="-1",
                role="menu",
                aria_labelledby=triggerid,
                data_floating_menu_direction=direction,
                id=menuid,
            ),
            **attributes,
        )
        if menuname is not None:
            self[0].insert(0, hg.SPAN(menuname, _class="bx--assistive-text"))



import os

import htmlgenerator as hg
from django.contrib.staticfiles import finders
from django.utils.html import mark_safe

RAW_ICON_BASE_PATH = "design/carbon_design/icons/flat/raw_32/"


class Icon(hg.SVG):

    """Insert the SVG for a carbon icon.
    See https://www.carbondesignsystem.com/guidelines/icons/library for a list of all icons.
    In order to see the name which should be passed to this template tag, click on "Download SVG"
    for an icon and use the filename without the attribte, e.g. "thunderstorm--severe"."""

    ICONS = None

    def __init__(
        self,
        name,
        size=None,
        **attributes,
    ):
        if Icon.ICONS is None:
            Icon.ICONS = loadicons(RAW_ICON_BASE_PATH)
        attributes["viewBox"] = "0 0 32 32"
        attributes["preserveAspectRatio"] = "xMidYMid meet"
        attributes["focusable"] = "false"
        attributes["style"] = attributes.get("style", "") + " will-change: transform;"
        if size is None:
            attributes["width"] = "32"
            attributes["height"] = "32"
        else:
            attributes["width"] = size
            attributes["height"] = size
        self.name = name
        super().__init__(
            hg.F(lambda c: Icon.ICONS[hg.resolve_lazy(name, c)]), **attributes
        )


def loadicons(basepath):
    absolute_path = finders.find(basepath)
    icons = {}

    for name in os.listdir(absolute_path):
        iconpath = os.path.join(absolute_path, name)
        if os.path.isfile(iconpath):
            with open(iconpath) as f:
                icons[name.replace(".svg", "")] = mark_safe(f.read())
    return icons



import json

import htmlgenerator as hg
from django import forms
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core import checks
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from djangoql.exceptions import DjangoQLError
from djangoql.queryset import apply_search
from djangoql.schema import DjangoQLSchema
from djangoql.serializers import DjangoQLSchemaSerializer

from bread import layout


class QueryValue:
    def __init__(self, queryset, raw):
        self.queryset = queryset
        self.raw = raw

    def __str__(self):
        return self.raw


class QuerySetDescriptor:
    def __init__(self, field):
        self.field = field

    def __get__(self, instance=None, owner=None):
        if instance is None:
            return self
        if self.field.name not in instance.__dict__:
            instance.refresh_from_db(fields=[self.field.name])
        value = instance.__dict__[self.field.name]
        model = getattr(instance, self.field.modelfieldname, None)
        if model and model.model_class():
            try:
                return parsequeryexpression(model.model_class().objects, value)
            except ValidationError:
                pass
        return QueryValue(None, value)

    def __set__(self, instance, value):
        instance.__dict__[self.field.name] = self.field.get_clean_value(value)


class QuerysetField(models.TextField):
    descriptor_class = QuerySetDescriptor

    def __init__(self, *args, modelfieldname, **kwargs):
        self.modelfieldname = modelfieldname
        kwargs["blank"] = True
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        return name, path, args, {"modelfieldname": self.modelfieldname, **kwargs}

    def contribute_to_class(self, cls, name):
        super().contribute_to_class(cls, name)
        setattr(cls, self.name, self.descriptor_class(self))

    def get_clean_value(self, value):
        return str(value)

    def get_prep_value(self, value):
        super().get_prep_value(self.get_clean_value(value))

    def value_to_string(self, obj):
        return self.get_prep_value(self.value_from_object(obj))

    def get_db_prep_value(self, value, connection, prepared=False):
        return self.get_clean_value(value)

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_field_name(),
        ]

    def _check_field_name(self):
        try:
            self.model._meta.get_field(self.modelfieldname)
        except FieldDoesNotExist:
            return [
                checks.Error(
                    "The QuerysetField modelfieldname references the "
                    "nonexistent field '%s'." % self.modelfieldname,
                    obj=self,
                    id="queryfield.E001",
                )
            ]
        else:
            return []

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        model = getattr(model_instance, self.modelfieldname)
        if not model or not model.model_class():
            raise ValidationError(_("Invalid model '%s' selected") % model)
        parsequeryexpression(
            getattr(model_instance, self.modelfieldname).model_class().objects, value
        )

    def formfield(self, **kwargs):
        return super().formfield(**{**kwargs, "form_class": QuerysetFormField})


class QuerysetFormField(forms.CharField):
    pass


class QuerysetFormWidget(layout.forms.widgets.Textarea):
    django_widget = QuerysetFormField

    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        inputelement_attrs=None,
        boundfield=None,
        **attributes
    ):
        super().__init__(
            label=label,
            help_text=help_text,
            errors=errors,
            inputelement_attrs=inputelement_attrs,
            boundfield=boundfield,
            **attributes
        )

        def introspections(context):
            return json.dumps(
                DjangoQLSchemaSerializer().serialize(
                    DjangoQLSchema(
                        hg.resolve_lazy(boundfield, context).value().queryset.model
                    )
                )
            )

        if boundfield:
            self.append(
                hg.SCRIPT(
                    hg.format(
                        """
                        document.addEventListener("DOMContentLoaded", () => DjangoQL.DOMReady(function () {{
                            new DjangoQL({{
                            introspections: {},
                            selector: 'textarea[name={}]',
                            syntaxHelp: '{}',
                            autoResize: false
                            }});
                        }}));
                    """,
                        hg.F(introspections),
                        inputelement_attrs.get("name"),
                        reverse("reporthelp"),
                        autoescape=False,
                    ),
                )
            )
            self.append(
                hg.LINK(
                    rel="stylesheet",
                    type="text/css",
                    href=staticfiles_storage.url("djangoql/css/completion.css"),
                )
            )
            self.append(
                hg.SCRIPT(src=staticfiles_storage.url("djangoql/js/completion.js"))
            )


def parsequeryexpression(basequeryset, expression):
    if not expression:
        return QueryValue(basequeryset.all(), expression)
    try:
        return QueryValue(apply_search(basequeryset, expression), expression)
    except DjangoQLError as e:
        raise ValidationError(str(e))



import htmlgenerator as hg
from django.conf import settings
from django.utils import formats
from django.utils.translation import gettext_lazy as _

from ..button import Button

REQUIRED_LABEL = getattr(settings, "REQUIRED_LABEL", " *")


class Submit(hg.DIV):
    def __init__(self, label=None, **attributes):
        super().__init__(
            Button(label or _("Save"), type="submit"),
            _class="bx--form-item",
            **attributes,
        )


class Label(hg.If):
    def __init__(self, label, _for=None, required=None, disabled=None, **kwargs):
        self.label = label
        super().__init__(
            label,
            hg.LABEL(
                label,
                hg.If(required, REQUIRED_LABEL),
                _for=_for,
                _class=hg.BaseElement(
                    "bx--label",
                    hg.If(disabled, " bx--label--disabled"),
                ),
                **kwargs,
            ),
        )


class HelpText(hg.If):
    def __init__(self, helptext, disabled=False):
        super().__init__(
            helptext,
            hg.DIV(
                helptext,
                _class=hg.BaseElement(
                    "bx--form__helper-text",
                    hg.If(disabled, " bx--form__helper-text--disabled"),
                ),
            ),
        )


class ErrorList(hg.If):
    def __init__(self, errors):
        super().__init__(
            errors,
            hg.DIV(
                hg.UL(hg.Iterator(errors or (), "error", hg.LI(hg.C("error")))),
                _class="bx--form-requirement",
            ),
        )


PHP_FORMAT_CHARACTERS = "dDjlNSwzWFmMntLoYyaABgGhHisuveIOPTZcrU"

# maps python format characters to PHP format characters
# hint: the #-sign of the python formatters is ommited for the keys
LETTER_MAPPING = {
    "a": "D",
    "A": "l",
    "w": "w",
    "d": "d",
    "b": "M",
    "B": "F",
    "m": "m",
    "y": "y",
    "Y": "Y",
    "H": "H",
    "I": "h",
    "p": "A",
    "M": "i",
    "S": "s",
    "f": "u",
    "z": "O",
    "Z": "e",
    "%": "%",
    "G": "o",
    "u": "N",
    "V": "W",
}


def to_php_formatstr(formatstr, format_key=None):
    """Maps a python datetime format string to a PHP format string

    This function is usefull because a lot of template/front-end code uses
    the PHP-format string, e.g. the date-filter of django.
    The following format specifiers are currently not supported:
    "%j", "%U", "%W", "%c", "%x", "%X"
    If formatstr is none, the format_key will be used to lookup default django format strings.
    """
    formatstr = " " + (formatstr or formats.get_format(format_key)[0])
    ret = []
    for i, c in enumerate(formatstr):
        if c == "%" or i == 0:
            continue

        if formatstr[i - 1] == "%":
            if c not in LETTER_MAPPING:
                raise ValueError(
                    f"Format letter %{c}  in format string {formatstr[1:]} is currently not supported to be converted to a PHP date-formatting string"
                )
            ret.append(LETTER_MAPPING[c])
        else:
            if c in PHP_FORMAT_CHARACTERS:
                ret.append("\\" + c)
            else:
                ret.append(c)
    return "".join(ret)



from .add import *  # noqa
from .auth import *  # noqa
from .browse import *  # noqa
from .delete import *  # noqa
from .edit import *  # noqa
from .read import *  # noqa
from .util import *  # noqa
from .wizard import *  # noqa



import htmlgenerator as hg
from django.urls import path
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView

from bread import menu, views
from bread.utils import urls
from bread.utils.links import Link, ModelHref
from bread.views.browse import delete
from bread.views.edit import bulkcopy, generate_copyview

from .models import Report
from .views import EditView, ReadView, exceldownload

urlpatterns = [
    *urls.default_model_paths(
        Report,
        browseview=views.BrowseView._with(
            columns=["name", "created"],
            rowclickaction=views.BrowseView.gen_rowclickaction("read"),
            bulkactions=[
                views.browse.BulkAction(
                    "delete", label=_("Delete"), iconname="trash-can", action=delete
                ),
                views.browse.BulkAction(
                    "copy",
                    label=_("Copy"),
                    iconname="copy",
                    action=lambda request, qs: bulkcopy(
                        request, qs, labelfield="name", copy_related_fields=("columns",)
                    ),
                ),
            ],
            rowactions=[
                Link(
                    href=ModelHref.from_object(hg.C("row"), "edit"),
                    iconname="edit",
                    label=_("Edit"),
                ),
                Link(
                    href=ModelHref.from_object(hg.C("row"), "excel"),
                    iconname="document",
                    label=_("Excel"),
                ),
            ],
        ),
        addview=views.AddView._with(
            fields=["name", "model"],
            get_success_url=lambda s: urls.reverse_model(
                Report, "edit", kwargs={"pk": s.object.pk}
            ),
        ),
        editview=EditView,
        readview=ReadView,
        copyview=generate_copyview(
            Report, labelfield="name", copy_related_fields=("columns",)
        ),
    ),
    urls.autopath(
        exceldownload,
        urls.model_urlname(Report, "excel"),
    ),
    path(
        "reporthelp/",
        TemplateView.as_view(template_name="djangoql/syntax_help.html"),
        name="reporthelp",
    ),
]

menu.registeritem(
    menu.Item(
        Link(
            urls.reverse_model(Report, "browse"),
            label=_("Reports"),
            iconname="document",
        ),
        menu.Group(_("Reports"), iconname="document"),
    )
)



from django.apps import apps
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Print a workflow diagram in the dot language"

    def add_arguments(self, parser):
        parser.add_argument("workflow")

    def handle(self, *args, **options):
        Workflow = apps.get_model(options["workflow"])
        print(Workflow.as_svg())



import htmlgenerator as hg


class Grid(hg.DIV):
    """
    Grid is a system uses a series of rows, and columns to layout and align content.
    Grid was built with flexbox and is fully responsive based on the IBM Design's 2x Grid System.

    Grid is usually combined with Row and Col, but not Grid itself alone.
    """

    MODES = ("narrow", "condensed", "full-width")

    def __init__(self, *children, gridmode="full-width", gutter=True, **attributes):
        attributes["_class"] = attributes.get("_class", "") + " bx--grid"
        if gridmode is not None:
            if gridmode not in Grid.MODES:
                raise ValueError(f"argument 'gridmode' must be one of {Grid.MODES}")
            attributes["_class"] += f" bx--grid--{gridmode}"
        if not gutter:
            attributes["_class"] += " bx--no-gutter"
        super().__init__(*children, **attributes)


class Row(hg.DIV):
    """
    Row is a container of a series of columns. Row itself can act as an area of columns,
    or as a part of Grid. In other words, Row is not required to be inside a Grid.
    """

    def __init__(self, *children, gridmode=None, **attributes):
        attributes["_class"] = hg.BaseElement(attributes.get("_class", ""), " bx--row")
        if gridmode is not None:
            attributes["_class"] = hg.BaseElement(
                f" bx--row--{gridmode}", attributes["_class"]
            )
        super().__init__(*children, **attributes)


class Col(hg.DIV):
    """
    Col is a container used within Rows. Col can come with both flexible width and fixed one
    by assigning the width property.
    """

    def __init__(self, *children, breakpoint="lg", width=None, **attributes):
        """
        breakpoint: Can be one of "sm", "md", "lg", "xlg", "max"
        """
        colclass = "bx--col"
        if width is not None:
            colclass += f"-{breakpoint}-{width}"
        attributes["_class"] = attributes.get("_class", "") + f" {colclass}"
        super().__init__(*children, **attributes)



import htmlgenerator as hg
from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _

from bread import layout, views

from .models import DocumentTemplate


class DocumentTemplateEditView(views.EditView):
    def get_layout(self):
        modelclass = self.object.model.model_class()
        if modelclass is None:
            return layout.notification.InlineNotification(
                _("Error"),
                f"Model '{self.object.model}' does no longer exist.",
                kind="error",
            )
        column_helper = layout.modal.Modal(
            _("Field explorer"), layout.fieldexplorer.field_help(modelclass), size="lg"
        )

        only_template, only_definition = None, None
        warnings = hg.BaseElement()
        try:
            only_template, only_definition = self.object.missing_variables()
        except Exception as e:
            warnings.append(
                layout.notification.InlineNotification(
                    _("Error in template: "),
                    str(e),
                    kind="error",
                )
            )
        if only_template:
            warnings.append(
                layout.notification.InlineNotification(
                    _("Variables in document but not defined below: "),
                    f"{', '.join(only_template)}",
                    kind="warning",
                )
            )
        if only_definition:
            warnings.append(
                layout.notification.InlineNotification(
                    _("Variables defined below but not used in document: "),
                    f"{', '.join(only_definition)}",
                    kind="warning",
                )
            )

        F = layout.forms.FormField
        fieldstable = layout.forms.FormsetField.as_datatable(
            "variables",
            [
                "name",
                "value",
                F(
                    "template",
                    no_wrapper=True,
                    no_label=True,
                    no_helptext=True,
                    inputelement_attrs={"rows": 1},
                ),
            ],
            formsetfield_kwargs={"extra": 1},
        )
        fieldstable[0][1].insert(
            0,
            layout.button.Button(
                _("Help"),
                buttontype="ghost",
                **column_helper.openerattributes,
            ),
        )

        ret = hg.BaseElement(
            views.header(),
            warnings,
            layout.forms.Form(
                hg.C("form"),
                F("name"),
                F("file"),
                fieldstable,
                column_helper,
                layout.forms.helpers.Submit(),
            ),
        )
        return ret

    def get_success_url(self):
        return self.request.get_full_path()


def generate_document(request, pk: int, object_pk: int):
    template = DocumentTemplate.objects.get(pk=pk)
    object = template.model.get_object_for_this_type(pk=object_pk)
    response = HttpResponse(
        template.render_with(object),
        content_type="application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    )
    response[
        "Content-Disposition"
    ] = f'attachment; filename="{template.name}_{str(object).replace(" ", "-")}.docx"'
    return response



import warnings
from typing import List, Optional, Type, Union

import htmlgenerator as hg
from django import forms

from .helpers import ErrorList, HelpText, Label
from .widgets import BaseWidget, HiddenInput, TextInput

DEFAULT_FORM_CONTEXTNAME = "__bread_form"
DEFAULT_FORMSET_CONTEXTNAME = "__bread_formset_form"


class FormFieldMarker(hg.BaseElement):
    # Internal helper class to mark form fields inside a render tree
    # so that the fields an be automatically extracted from it in to
    # generate a django form class, see bread.forms.forms
    def __init__(self, fieldname, field):
        self.fieldname = fieldname
        super().__init__(field)


def generate_widget_element(
    fieldname: str = None,  # required to derive the widget from a django form field
    form: Union[
        forms.Form, hg.Lazy, str
    ] = DEFAULT_FORM_CONTEXTNAME,  # required to derive the widget from a django form field
    no_wrapper: bool = False,  # wrapper produces less dense layout, from carbon styles
    no_label: bool = False,
    no_helptext: bool = False,
    show_hidden_initial: bool = False,  # required in special cases to add an initial value
    #
    #
    # --------------------------------------------------------------------------
    # parameters which are normally not required, when using a django form field
    # but can be filled in to create form fields independently from django form fields or
    # manually overriding values from the form field
    widgetclass: Optional[
        Union[Type[BaseWidget], hg.Lazy]
    ] = None,  # normally be taken from the django form field, will be carbon-ized
    label: Union[
        str, hg.BaseElement
    ] = None,  # normally be taken from the django form field, will be carbon-ized
    help_text: Union[
        str, hg.BaseElement
    ] = None,  # normally be taken from the django form field, will be carbon-ized
    errors: Optional[
        List[str]
    ] = None,  # normally be taken from the django form field, will be carbon-ized
    inputelement_attrs: Optional[
        Union[dict, hg.Lazy]
    ] = None,  # normally be taken from the django form field, will be carbon-ized
    **attributes,
) -> FormFieldMarker:
    """
    Function to produce a carbon design based form field widget which is
    compatible with Django forms and based on htmlgenerator.
    """

    hidden = None
    if show_hidden_initial:
        hidden = generate_widget_element(
            fieldname=fieldname,
            form=form,
            inputelement_attrs=inputelement_attrs,
            widgetclass=HiddenInput,
            no_wrapper=True,
            no_label=True,
            no_helptext=True,
            show_hidden_initial=False,
            **attributes,
        )

    inputelement_attrs = inputelement_attrs or {}
    boundfield = None

    # warnings for deprecated API usage
    if "widgetattributes" in attributes:
        warnings.warn(
            "FormField does no longer support the parameter 'widgetattributes'. "
            "The parameter 'inputelement_attrs' serves the same purpose'"
        )
    if "elementattributes" in attributes:
        warnings.warn(
            "FormField does no longer support the parameter 'elementattributes'. "
            "attributes can now be directly passed as kwargs."
        )

    # check if this field will be used with a django form if yes, derive the
    # according values lazyly from the context
    if fieldname is not None and form is not None:
        if isinstance(form, str):
            form = hg.C(form)

        label = label or form[fieldname].label
        help_text = help_text or form.fields[fieldname].help_text
        errors = errors or form[fieldname].errors

        # do this to preserve the original inputelement_attrs in the
        # buildattribs scope
        orig_inputattribs = inputelement_attrs

        def buildattribs(context):
            realform = hg.resolve_lazy(form, context)
            id = None
            if realform[fieldname].auto_id and "id" not in orig_inputattribs:
                id = (
                    realform[fieldname].html_initial_id
                    if show_hidden_initial
                    else realform[fieldname].auto_id
                )
            return {
                "id": id,
                "name": realform[fieldname].html_initial_name
                if show_hidden_initial
                else realform[fieldname].html_name,
                "value": realform[fieldname].value(),
                **realform[fieldname].build_widget_attrs({}),
                **realform[fieldname].field.widget.attrs,
                **orig_inputattribs,
            }

        inputelement_attrs = hg.F(buildattribs)
        labelfor = form[fieldname].id_for_label
        boundfield = form[fieldname]
    else:
        labelfor = inputelement_attrs.get("id")

    # helper elements
    label = Label(
        label,
        required=inputelement_attrs.get("required"),
        disabled=inputelement_attrs.get("disabled"),
        _for=labelfor,
    )
    help_text = HelpText(help_text, disabled=inputelement_attrs.get("disabled"))
    errors = ErrorList(errors)

    # instantiate field (might create a lazy element when using _guess_widget)
    widgetclass = _guess_widget(fieldname, form, widgetclass)
    ret = widgetclass(
        label=None if no_label else label,
        help_text=None if no_helptext else help_text,
        errors=errors,
        inputelement_attrs=inputelement_attrs,
        boundfield=boundfield,
        **attributes,
    )
    if show_hidden_initial:
        ret = hg.BaseElement(ret, hidden)
    if not no_wrapper:
        ret = hg.If(
            hg.F(
                lambda c: isinstance(
                    hg.resolve_lazy(boundfield, c).field.widget, forms.HiddenInput
                )
            ),
            ret,
            ret.with_fieldwrapper(),
        )
    return FormFieldMarker(fieldname, ret)


# Using this alias we can prevent a huge refactoring across multiple repos This
# is slightly inconsistent with the default naming scheme of python where camel
# case denotes not a function but a class
# TODO: maybe refactor Formfield to be formfield
FormField = generate_widget_element


def _guess_widget(fieldname, form, suggested_widgetclass) -> hg.Lazy:
    widget_map: dict = {}
    for cls in _all_subclasses(BaseWidget):
        if cls.django_widget not in widget_map:
            widget_map[cls.django_widget] = []
        widget_map[cls.django_widget].append(cls)

    def wrapper(context):
        realform = hg.resolve_lazy(form, context)
        widgetclass = type(realform[fieldname].field.widget)
        fieldclass = type(realform[fieldname].field)

        # Hidden widgets have highest priority
        if issubclass(widgetclass, forms.HiddenInput):
            return HiddenInput
        # Manually passed widgets have second priority
        if suggested_widgetclass is not None:
            return suggested_widgetclass

        # Automated detection via django-bread-widget-mapp have lowest priority
        if fieldclass in widget_map:
            return widget_map[fieldclass][0]
        if widgetclass in widget_map:
            return widget_map[widgetclass][0]

        # Fallback for unknown widgets
        warnings.warn(
            f"Form field {type(realform).__name__}.{fieldname} ({fieldclass}) uses widget {widgetclass} but "
            "bread has no implementation, default to TextInput"
        )
        return TextInput

    return hg.F(wrapper)


def _all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in _all_subclasses(c)]
    )



import htmlgenerator as hg
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils.html import strip_tags
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _

from .components.notification import ToastNotification
from .components.shell_header import ShellHeader
from .components.sidenav import SideNav

_static = staticfiles_storage.url


def default_page_layout(menu, *content):
    return hg.HTML(
        hg.HEAD(
            hg.TITLE(
                hg.F(
                    lambda c: strip_tags(
                        hg.render(
                            hg.BaseElement(c.get("pagetitle", c.get("PLATFORMNAME"))), c
                        )
                    )
                    + " | "
                    + strip_tags(c.get("PLATFORMNAME"))
                )
            ),
            hg.LINK(rel="shortcut icon", href=_static("logo.png")),
            hg.LINK(
                rel="stylesheet",
                type="text/css",
                href=hg.If(
                    settings.DEBUG,
                    _static("css/bread-main.css"),  # generate with "make css"
                    _static("css/bread-main.min.css"),  # generate with "make css"
                ),
                media="all",
            ),
        ),
        hg.BODY(
            ShellHeader(
                hg.C("PLATFORMNAME"),
                hg.C("COMPANYNAME"),
                searchbar=hg.If(
                    hg.C("request.user.is_authenticated"),
                    hg.C("SEARCHBAR"),
                    "",
                ),
            ),
            hg.If(
                hg.C("request.user.is_authenticated"),
                SideNav(menu),
            ),
            hg.DIV(
                hg.Iterator(
                    hg.C("messages"),
                    "message",
                    ToastNotification(
                        message=hg.F(lambda c: _(c["message"].tags.capitalize())),
                        details=hg.C("message.message"),
                        kind=hg.C("message.level_tag"),
                        hidetimestamp=True,
                        autoremove=5.0,
                    ),
                ),
                style="position: fixed; right: 0; z-index: 999",
            ),
            hg.DIV(*content, _class="bx--content"),
            hg.If(
                settings.DEBUG,
                hg.BaseElement(
                    hg.SCRIPT(src=_static("js/bliss.js")),
                    hg.SCRIPT(src=_static("js/htmx.js")),
                    hg.SCRIPT(src=_static("js/main.js")),
                    hg.SCRIPT(
                        src=_static("design/carbon_design/js/carbon-components.js")
                    ),
                ),
                hg.SCRIPT(src=_static("js/bread.min.js")),  # generate with "make js"
            ),
            hg.SCRIPT("CarbonComponents.watch(document);"),
        ),
        doctype=True,
        _class="no-js",
        lang=get_language(),
    )



import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from .button import Button


def field_help(model, max_depth=4):
    return hg.BaseElement(
        hg.DIV(
            hg.format(_("Base model: ")),
            model._meta.verbose_name,
            style="margin-top: 2rem",
        ),
        hg.DIV(
            _("Click "),
            hg.SPAN(
                " ... ", style="width: 1rem; height: 1rem; background-color: lightgray;"
            ),
            " ",
            _(" to copy the accessor"),
            style="margin-top: 2rem",
        ),
        hg.DIV(_("Fields"), style="margin-top: 2rem"),
        hg.DIV(
            get_field_list(
                model,
                max_depth,
                list(
                    set([model, model.__mro__[-3], *_all_subclasses(model.__mro__[-3])])
                ),
                display="block",
            ),
            style="line-height: 2rem",
        ),
    )


def get_field_list(model, depth, excludemodels, display="none", parent_accessor=[]):
    fields = {}
    for f in model._meta.get_fields():
        if not f.concrete and (f.one_to_many or f.many_to_many):
            continue
        fields[f] = hg.DIV(
            hg.SPAN(
                ".".join(parent_accessor + [_field_attname(f)]),
                style="display:none;"
                "position: absolute;"
                "top: -2rem;"
                "background-color: white;"
                "padding: 0.25rem",
            ),
            hg.SPAN(
                " ... ",
                style="width: 1rem; height: 1rem; background-color: lightgray; cursor: pointer;",
                onmouseover="this.previousElementSibling.style.display = 'block'",
                onmouseout="this.previousElementSibling.style.display = 'none'",
                onclick="navigator.clipboard.writeText(this.previousElementSibling.innerText);",
            ),
            " ",
            hg.SPAN(
                hg.SPAN(_field_attname(f), style="font-weight: 700"),
                " ",
                _field_type_repr(f),
            ),
            style="position: relative",
        )
        if (
            f.related_model
            and not f.many_to_many
            and f.related_model not in excludemodels
            and depth > 0
        ):
            fields[f].append(
                Button(
                    buttontype="ghost",
                    icon="chevron--down",
                    onclick=_js_toggle_display("this.nextElementSibling"),
                )
            )
            fields[f].append(
                get_field_list(
                    f.related_model,
                    depth - 1,
                    excludemodels + [model],
                    parent_accessor=parent_accessor + [_field_attname(f)],
                )
            )
    return hg.UL(
        hg.Iterator(
            sorted(
                fields.items(),
                key=lambda k: ("1" if k[0].related_model else "0")
                + _field_attname(k[0]),
            ),
            "field",
            hg.LI(hg.C("field.1")),
        ),
        style=f"margin-left: 2rem; display: {display}",
    )


def _field_type_repr(field):
    if field.related_model:
        if field.one_to_one or field.many_to_one:
            return _(' reference to "%s"') % field.related_model._meta.verbose_name
        if field.one_to_many or field.many_to_many:
            return (
                _(' reference to multiple "%s"')
                % field.related_model._meta.verbose_name_plural
            )
    return f"{type(field).__name__}({field.verbose_name})"


def _field_attname(field):
    attname = field.name
    if hasattr(field, "attname"):
        attname = field.attname
    elif hasattr(field, "get_accessor_name"):
        attname = field.get_accessor_name()
    if attname.endswith("_id"):
        attname = attname[:-3]
    return attname


def _js_toggle_display(element_accessor):
    return (
        f"{element_accessor}.style.display = "
        f"{element_accessor}.style.display == 'none' ? 'block': 'none'"
    )


def _all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in _all_subclasses(c)]
    )



import typing

import htmlgenerator as hg
from django.urls import path

from bread import layout
from bread.utils import quickregister
from bread.views import AddView, EditView

from .models import DataChangeTrigger, DateFieldTrigger, SendEmail

urlpatterns: typing.List[path] = []
quickregister(
    urlpatterns,
    DataChangeTrigger,
    editview=EditView._with(
        fields=[
            hg.H4(
                layout.ObjectFieldLabel("model"), ": ", layout.ObjectFieldValue("model")
            ),
            "action",
            "type",
            "filter",
            "enable",
        ]
    ),
    addview=AddView._with(fields=["description", "model", "type", "action"]),
)
quickregister(
    urlpatterns,
    DateFieldTrigger,
    editview=EditView._with(
        fields=[
            hg.H4(
                layout.ObjectFieldLabel("model"), ": ", layout.ObjectFieldValue("model")
            ),
            "action",
            "offset_type",
            "offset_amount",
            "field",
            "filter",
            "enable",
        ]
    ),
    addview=AddView._with(
        fields=[
            "description",
            "model",
            "action",
            "offset_type",
            "offset_amount",
            "field",
        ]
    ),
)
quickregister(urlpatterns, SendEmail)



from typing import Iterator, NamedTuple

import htmlgenerator as hg
from django.core.paginator import Paginator
from django.utils.translation import gettext_lazy as _

from bread.utils.urls import link_with_urlparameters

from ..utils import aslink_attributes
from .forms.widgets import Select
from .icon import Icon


class PaginationConfig(NamedTuple):
    paginator: Paginator
    items_per_page_options: Iterator
    page_urlparameter: str = (
        "page"  # URL parameter which holds value for current page selection
    )
    itemsperpage_urlparameter: str = (
        "itemsperpage"  # URL parameter which selects items per page
    )


class Pagination(hg.DIV):
    def __init__(
        self,
        paginator,
        items_per_page_options,
        page_urlparameter="page",
        itemsperpage_urlparameter="itemsperpage",
        **kwargs,
    ):
        select1_id = hg.html_id(self)
        select2_id = select1_id + "-n"
        kwargs["_class"] = kwargs.get("_class", "") + " bx--pagination"
        kwargs["data_pagination"] = True
        super().__init__(
            hg.DIV(
                hg.LABEL(
                    _("Items per page"),
                    ":",
                    _class="bx--pagination__text",
                    _for=select1_id,
                ),
                Select(
                    choices=[
                        (
                            linkwithitemsperpage(
                                itemsperpage_urlparameter,
                                page_urlparameter,
                                itemsperpage=i,
                            ),
                            _("All") if i == -1 else i,
                        )
                        for i in items_per_page_options
                    ],
                    inline=True,
                    inputelement_attrs={
                        "data_items_per_page": True,
                        "onchange": "document.location = this.value",
                        "onauxclick": "window.open(this.value, '_blank')",
                        "value": linkwithitemsperpage(
                            itemsperpage_urlparameter,
                            page_urlparameter,
                        ),
                    },
                    _class="bx--select__item-count",
                ),
                hg.SPAN(
                    hg.SPAN(
                        hg.format(
                            "{} - {}",
                            get_page(paginator, page_urlparameter).start_index(),
                            get_page(paginator, page_urlparameter).end_index(),
                        ),
                        data_displayed_item_range=True,
                    ),
                    " ",
                    _("of"),
                    " ",
                    hg.SPAN(
                        hg.format(" {} ", paginator.count),
                        data_total_items=True,
                    ),
                    " ",
                    _("items"),
                    _class="bx--pagination__text",
                ),
                _class="bx--pagination__left",
            ),
            hg.DIV(
                Select(
                    choices=[
                        (linktopage(page_urlparameter, i), i)
                        for i in paginator.page_range
                    ],
                    inline=True,
                    inputelement_attrs={
                        "data_page_number_input": True,
                        "onchange": "document.location = this.value",
                        "onauxclick": "window.open(this.value, '_blank')",
                        "value": linktopage(
                            page_urlparameter,
                            hg.C("request").GET.get(page_urlparameter, "1"),
                        ),
                    },
                    _class="bx--select__page-number",
                ),
                hg.LABEL(
                    _("of"),
                    " ",
                    paginator.num_pages,
                    " ",
                    _("pages"),
                    _class="bx--pagination__text",
                    _for=select2_id,
                ),
                hg.BUTTON(
                    Icon("caret--left", size=20, _class="bx--pagination__nav-arrow"),
                    _class="bx--pagination__button bx--pagination__button--backward",
                    tabindex="0",
                    type="button",
                    disabled=hg.F(
                        lambda c: not paginator.get_page(
                            int(c["request"].GET.get(page_urlparameter, "1"))
                        ).has_previous()
                    ),
                    data_page_backward=True,
                    **aslink_attributes(
                        linktorelativepage(
                            page_urlparameter,
                            -1,
                            paginator.num_pages,
                        )
                    ),
                ),
                hg.BUTTON(
                    Icon("caret--right", size=20, _class="bx--pagination__nav-arrow"),
                    _class="bx--pagination__button bx--pagination__button--forward",
                    tabindex="0",
                    type="button",
                    disabled=hg.F(
                        lambda c: not paginator.get_page(
                            int(c["request"].GET.get(page_urlparameter, "1"))
                        ).has_next()
                    ),
                    data_page_forward=True,
                    **aslink_attributes(
                        linktorelativepage(
                            page_urlparameter,
                            1,
                            paginator.num_pages,
                        )
                    ),
                ),
                _class="bx--pagination__right",
            ),
            **kwargs,
        )

    @classmethod
    def from_config(cls, pagination_config):
        return cls(
            paginator=pagination_config.paginator,
            items_per_page_options=pagination_config.items_per_page_options,
            page_urlparameter=pagination_config.page_urlparameter,
            itemsperpage_urlparameter=pagination_config.itemsperpage_urlparameter,
        )


def linktopage(page_urlparameter, page):
    return hg.F(
        lambda c: link_with_urlparameters(
            c["request"], **{page_urlparameter: hg.resolve_lazy(page, c)}
        )
    )


def linktorelativepage(page_urlparameter, direction, maxnum):
    """direction: number of pages to jump, e.g. -1 or 1"""
    return hg.F(
        lambda c: link_with_urlparameters(
            c["request"],
            **{
                page_urlparameter: int(c["request"].GET.get(page_urlparameter, "1"))
                + hg.resolve_lazy(direction, c)
            },
        )
    )


def linkwithitemsperpage(
    itemsperpage_urlparameter, page_urlparameter, itemsperpage=None
):
    return hg.F(
        lambda c: link_with_urlparameters(
            c["request"],
            **{
                itemsperpage_urlparameter: itemsperpage
                or c["request"].GET.get(itemsperpage_urlparameter),
                page_urlparameter: None,
            },
        )
    )


def get_page(paginator, page_urlparameter):
    def wrapper(context):
        return paginator.get_page(
            int(context["request"].GET.get(page_urlparameter, "1"))
        )

    return hg.F(wrapper)



import typing

import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from .icon import Icon
from .loading import Loading


class SearchBackendConfig(typing.NamedTuple):
    """Describes an endpoint for auto-complete searches"""

    url: typing.Any
    result_selector: str = ""
    result_label_selector: str = ""
    result_value_selector: str = ""
    query_parameter: str = "q"


class Search(hg.DIV):
    def __init__(
        self,
        size="xl",
        placeholder=None,
        widgetattributes=None,
        backend=None,
        resultcontainerid=None,
        show_result_container=True,
        resultcontainer_onload_js=None,
        disabled=False,
        **kwargs,
    ):
        """
        :param SearchBackendConfig backend: Where and how to get search results
        """
        kwargs["_class"] = kwargs.get("_class", "") + f" bx--search bx--search--{size}"
        kwargs["data_search"] = True
        kwargs["role"] = "search"
        width = kwargs.get("width", None)
        if width:
            kwargs["style"] = kwargs.get("style", "") + f"width:{width};"

        widgetattributes = {
            "id": "search__" + hg.html_id(self),
            "_class": "bx--search-input",
            "type": "text",
            "placeholder": placeholder or _("Search"),
            "autocomplete": "off",
            **(widgetattributes or {}),
        }
        if backend:
            if resultcontainerid is None:
                resultcontainerid = f"search-result-{hg.html_id((self, backend.url))}"
            widgetattributes["hx_get"] = backend.url
            widgetattributes["hx_trigger"] = "changed, click, keyup changed delay:500ms"
            widgetattributes["hx_target"] = hg.format("#{}", resultcontainerid)
            widgetattributes["hx_indicator"] = hg.format(
                "#{}-indicator", resultcontainerid
            )
            widgetattributes["name"] = backend.query_parameter

        self.close_button = _close_button(resultcontainerid, widgetattributes)

        super().__init__(
            hg.DIV(
                hg.LABEL(_("Search"), _class="bx--label", _for=widgetattributes["id"]),
                hg.INPUT(**widgetattributes),
                _search_icon(),
                self.close_button,
                hg.If(backend is not None, _loading_indicator(resultcontainerid)),
                **kwargs,
            ),
            hg.If(
                backend is not None and show_result_container,
                _result_container(resultcontainerid, resultcontainer_onload_js, width),
            ),
            style=hg.If(disabled, hg.BaseElement("display: none")),
        )


def _result_container(_id, onload_js, width="100%"):
    return hg.DIV(
        hg.DIV(
            id=_id,
            style="width: 100%; position: absolute; z-index: 999",
            onload=onload_js,
        ),
        style=f"width: {width}; position: relative",
    )


def _search_icon():
    return Icon(
        "search",
        size=16,
        _class="bx--search-magnifier",
        aria_hidden="true",
    )


def _loading_indicator(resultcontainerid):
    return hg.DIV(
        Loading(small=True),
        id=hg.format("{}-indicator", resultcontainerid),
        _class="htmx-indicator",
        style="position: absolute; right: 2rem",
    )


def _close_button(resultcontainerid, widgetattributes):
    kwargs = {
        "_class": hg.BaseElement(
            "bx--search-close",
            hg.If(widgetattributes.get("value"), None, " bx--search-close--hidden"),
        ),
        "title": _("Clear search input"),
        "aria_label": _("Clear search input"),
        "type": "button",
    }
    if resultcontainerid is not None:
        kwargs["onclick"] = hg.format(
            "document.getElementById('{}').innerHTML = '';", resultcontainerid
        )
    return hg.BUTTON(Icon("close", size=20, _class="bx--search-clear"), **kwargs)



import re
import subprocess  # nosec because all dealt with

import htmlgenerator as hg
from django.db import models
from django.utils import timezone
from django.utils.html import mark_safe

from .nodes import (
    Action,
    Decision,
    FlowFinal,
    Fork,
    Initial,
    Join,
    Merge,
    Node,
    WorkflowFinal,
)


class WorkflowBase(models.Model):
    """
    This class represents a workflow in the form of a UML Activity Diagram. In order to
    make the purpose of the system more clear the word "Workflow" is used in all places
    instead of "Activity".
    In order to persist state for an instance of a workflow Django models can be used.
    The diagram itself is defined in a class-variable of type dict named ``WORKFLOW``.
    The keys of ``WORKFLOW`` represent source nodes and the values the target nodes.
    Keys always need to be of type :py:class:`bread.contrib.acitivities.models.Node`.
    Values must be of type :py:class:`bread.contrib.acitivities.models.Node`.
    Exceptions are entries with a Fork node as key and entries with a Decision node as key.
    They require tuple(:py:class:`bread.contrib.acitivities.models.Node`) and of type dict{str: :py:class:`bread.contrib.acitivities.models.Node`} as values.
    """

    started = models.DateTimeField(auto_now_add=True, editable=False)
    completed = models.DateTimeField(null=True, editable=False)
    cancelled = models.DateTimeField(null=True, editable=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_workflow_state()

    @classmethod
    def workflowdiagram(cls):
        """
        Method which does the magic of converting a graph from simple dict format to
        full fledged activity diagram and do quite a bit of verification. Requires an
        attribute ``WORKFLOW`` of type ``dict``
        """

        if not hasattr(cls, "_GENERATED_DIAGRAM"):
            cls._GENERATED_DIAGRAM = generate_workflowdiagram(cls)
        return cls._GENERATED_DIAGRAM

    @classmethod
    def workflow_as_dot(cls, attrs=None):
        default_attrs = {"splines": "ortho"}
        default_attrs.update(attrs or {})
        dot = [
            f'digraph "{cls._meta.verbose_name}" {{n',
            *[f"{k}={v}" for k, v in default_attrs.items()],
            'graph[bgcolor="#ffffff00" ranksep=1]',
            "edge[arrowhead=open]",
        ]
        nodes = set()
        edges = []
        for source, target, choice in cls.workflowdiagram():
            nodes.add(source.as_dot())
            nodes.add(target.as_dot())
            edges.append(
                (
                    source,
                    target,
                    f'taillabel = "[{choice}]" labeldistance = 3.0' if choice else "",
                )
            )

        dot.extend(n for n in nodes)
        dot.extend(
            f"{id(n1)} -> {id(n2)}[{' '.join(attrs)}]" for n1, n2, *attrs in edges
        )

        dot.append("}")
        return "\n".join(dot)

    @classmethod
    def workflow_as_svg(cls, attrs=None):
        return dot2svg(cls.workflow_as_dot(attrs))

    def active_fields(self):
        ret = []
        for field in self._meta.get_fields():
            if isinstance(field, (Action, Decision)):
                if not field.done(self) and field.hasincoming(self):
                    ret.append(field.name)
        return ret

    def as_dot(self, attrs=None):
        def edgecolor(source, target, choice, instance):
            if isinstance(source, Decision):
                return (
                    "lightgreen"
                    if getattr(instance, source.name) == choice
                    else "black"
                )
            if source.done(instance) and target.done(instance):
                return "lightgreen"
            return "black"

        def nodecolor(node, instance):
            if node.done(instance):
                return "lightgreen"
            if node.hasincoming(instance):
                return "orange"
            return "black"

        default_attrs = {"splines": "ortho", "overlap": "true"}
        default_attrs.update(attrs or {})
        dot = [
            f'digraph "{self._meta.verbose_name}" {{\nranksep=2',
            *[f"{k}={v}" for k, v in default_attrs.items()],
            'graph[bgcolor="#ffffff00" ranksep=1]',
            "edge[arrowhead=open]",
        ]
        nodes = set()
        edges = []
        for source, target, choice in self.workflowdiagram():
            nodes.add(
                source.as_dot(
                    attrs={
                        "fillcolor": "lightgreen" if source.done(self) else "lightgrey",
                        "color": nodecolor(source, self),
                    }
                )
            )
            nodes.add(
                target.as_dot(
                    attrs={
                        "fillcolor": "lightgreen" if target.done(self) else "lightgrey",
                        "color": nodecolor(target, self),
                    }
                )
            )
            edges.append(
                (
                    source,
                    target,
                    f'taillabel = "[{choice}]" labeldistance = 3.0' if choice else "",
                    f"color={edgecolor(source, target, choice, self)}",
                    *(("headport=n",) if isinstance(target, Join) else ()),
                    *(("tailport=s",) if isinstance(target, Fork) else ()),
                )
            )

        dot.extend(n for n in nodes)
        dot.extend(
            f"{id(n1)} -> {id(n2)}[{' '.join(attrs)}]" for n1, n2, *attrs in edges
        )

        dot.append("}")
        return "\n".join(dot)

    def as_svg(self, attrs=None):
        return dot2svg(self.as_dot(attrs))

    @property
    def done(self):
        """Checks whether any Final nodes have been reached"""
        return any(
            isinstance(target, WorkflowFinal) and target.done(self)
            for source, target, choice in self.workflowdiagram()
        )

    def cancel(self, save=True):
        if self.completed:
            raise RuntimeError("Workflow has already completed, cannot cancel")
        self.cancelled = timezone.now()
        if save:
            self.save()

    def save(self, *args, **kwargs):
        if self.pk is not None:
            self.update_workflow_state(runactions=True)
        if not self.completed and not self.cancelled and self.done:
            self.completed = timezone.now()
        super().save(*args, **kwargs)
        self.update_workflow_state(runactions=True)
        super().save(*args, **kwargs)

    def update_workflow_state(self, runactions=False):
        # don't do anything on the workflow after it has been cancelled
        if self.cancelled:
            return
        state_changed = True
        while state_changed:
            nodequeue = [
                source
                for source, target, choice in self.workflowdiagram()
                if isinstance(source, Initial)
            ]
            state_changed = False
            while nodequeue:
                node = nodequeue.pop()
                if not node.done(self):
                    if isinstance(node, Action):
                        if (
                            node.hasincoming(self)
                            and not getattr(self, node.name)
                            and runactions
                        ):
                            actionresult = node.action(self)
                            if actionresult != getattr(self, node.name):
                                state_changed = True
                                setattr(self, node.name, actionresult)
                    if isinstance(node, Decision):
                        if any(n.done(self) for n, c in node.inputs):
                            decision = node.decide(self)
                            if decision != getattr(self, node.name):
                                state_changed = True
                                setattr(self, node.name, decision)
                for outputnode, choice in node.outputs:
                    nodequeue.insert(0, outputnode)

    class Meta:
        abstract = True


def dot2svg(dot: str):
    try:
        process = subprocess.run(  # nosec because fixed arguments and we take the risk of relative path
            ["dot", "-Tsvg"],
            input=dot.encode(),
            capture_output=True,
            check=True,
        )
        svg = process.stdout.decode()
        svg = re.sub('width="[0-9]*pt"', "", svg)
        svg = re.sub('height="[0-9]*pt"', "", svg)
        return mark_safe(svg)
    except subprocess.CalledProcessError as e:
        return mark_safe(
            hg.render(
                hg.DIV(
                    hg.DIV(
                        "Workflow diagram could not be generated, the error message was:"
                    ),
                    hg.DIV(hg.CODE(e)),
                    hg.DIV(hg.CODE(e.stderr.decode())),
                    hg.DIV(hg.PRE(hg.CODE(dot))),
                ),
                {},
            )
        )


def generate_workflowdiagram(model):  # noqa
    """Method which does the magic of converting a graph from simple dict format to
    full fledged activity diagram and do quite a bit of verification. Requires an attribute ``WORKFLOW`` of type ``dict``"""

    def allnodes(graph):
        return set(sum((list(i[:2]) for i in graph), []))

    assert hasattr(
        model, "WORKFLOW"
    ), f"class {model} needs to have an attribute 'WORKFLOW' with the workflowdefinition"  # nosec because only for API validation
    assert isinstance(  # nosec because only for API validation
        model.WORKFLOW, dict
    ), f"{model}.WORKFLOW needs to be of type {type(dict)}"
    graph = []  # [(source, target, choice)]

    # some type checking
    for node in allnodes(graph):
        assert isinstance(  # nosec because only for API validation
            node, (Node, dict, tuple, type(None))
        ), f"Node {node} if not of type {Node}"

    # flatten the WORKFLOW definition
    for source, target in model.WORKFLOW.items():
        if isinstance(target, tuple):
            graph.extend((source, node, None) for node in target)
        elif isinstance(target, dict):
            assert all(  # nosec because only for API validation
                t is not None for t in target.keys()
            ), "None is not an allowed key for choices"
            graph.extend((source, node, choice) for choice, node in target.items())
        else:
            graph.append((source, target, None))

    # insert initial and final nodes, mostly for activity diagram compliance
    for source, target, choice in list(graph):
        if source not in set(i[1] for i in graph) and not isinstance(source, Initial):
            graph.append((Initial(), source, None))
        if target not in set(i[0] for i in graph) and not isinstance(
            target, WorkflowFinal
        ):
            graph.append((target, WorkflowFinal(), None))
        if target is None:
            graph.append((target, FlowFinal(), None))

    # insert merge nodes
    removed_edges = []
    new_edges = []
    for node in allnodes(graph):
        inputs = [[s, choice] for s, t, choice in graph if t == node]
        if len(inputs) > 1 and not isinstance(node, (Join, Merge)):
            mergenode = Merge()
            new_edges.append((mergenode, node, None))
            for inputnode, choice in inputs:
                new_edges.append((inputnode, mergenode, choice))
                removed_edges.append((inputnode, node, choice))
    for edge in removed_edges:
        graph.remove(edge)
    graph.extend(new_edges)

    # insert fork nodes
    removed_edges = []
    new_edges = []
    for node in allnodes(graph):  # check if node needs to be a fork node
        outputs = [
            [t, choice] for s, t, choice in graph if s == node and choice is None
        ]
        if len(outputs) > 1:
            forknode = Fork()
            new_edges.append((node, forknode, None))
            for outputnode, choice in outputs:
                new_edges.append((forknode, outputnode, choice))
                removed_edges.append((node, outputnode, choice))
    for edge in removed_edges:
        graph.remove(edge)
    graph.extend(new_edges)

    # set inputs and outputs per node and verify all nodes
    for node in allnodes(graph):
        node.inputs = tuple(
            [(source, choice) for source, target, choice in graph if target == node]
        )
        node.outputs = tuple(
            [(target, choice) for source, target, choice in graph if source == node]
        )
        node._node_verify()
    return graph



# Generated by Django 3.2.12 on 2022-04-06 10:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0008_rename_name_reportcolumn_header'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportcolumn',
            name='sortingname',
            field=models.CharField(blank=True, help_text='Django sorting expression', max_length=255, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='reportcolumn',
            name='column',
            field=models.CharField(help_text="Value expression (see 'Help')", max_length=255, verbose_name='Column'),
        ),
        migrations.AlterField(
            model_name='reportcolumn',
            name='header',
            field=models.CharField(max_length=255, verbose_name='Header'),
        ),
    ]



from datetime import timedelta

from django.conf import settings

# DETERMINES HOW OFTEN WORKFLOWS WILL BE RUN TO
# TRIGGER AUTOMATED ACTIONS AND DECISIONS
WORKFLOW_BEAT = getattr(settings, "WORKFLOW_BEAT", timedelta(minutes=5))



import datetime

from celery import shared_task
from django.apps import AppConfig
from django.conf import settings
from django.utils import timezone

from bread.utils import get_concrete_instance

TRIGGER_PERIOD = getattr(settings, "TRIGGER_PERIOD", datetime.timedelta(hours=1))


class TriggersConfig(AppConfig):
    name = "bread.contrib.triggers"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self):

        from django.db.models.signals import post_delete, post_save

        from bread.utils.celery import RepeatedTask

        post_save.connect(save_handler, dispatch_uid="trigger_save")
        post_delete.connect(delete_handler, dispatch_uid="trigger_delete")

        shared_task(base=RepeatedTask, run_every=TRIGGER_PERIOD)(periodic_trigger)


def save_handler(sender, instance, created, **kwargs):
    datachange_trigger(sender, instance, "added" if created else "changed")


def delete_handler(sender, instance, **kwargs):
    datachange_trigger(sender, instance, "deleted")


def datachange_trigger(model, instance, type):
    from django.contrib.contenttypes.models import ContentType

    from .models import DataChangeTrigger

    model = ContentType.objects.filter(
        app_label=model._meta.app_label, model=model._meta.model_name
    )
    if model.exists():
        for trigger in DataChangeTrigger.objects.filter(
            model=model.first(), type=type, enable=True
        ):
            if trigger.filter.queryset.filter(pk=instance.pk).exists():
                # delay execution a bit as the trigger may run immediately even though
                # the current request has not finished (and therefore not commited to DB yet)
                run_action.apply_async(
                    (trigger.action.pk, instance._meta.label, instance.pk), countdown=5
                )


@shared_task
def run_action(action_pk, modelname, instance_pk):
    from django.apps import apps

    from .models import Action

    get_concrete_instance(Action.objects.get(pk=action_pk)).run(
        apps.get_model(modelname).objects.get(pk=instance_pk)
    )


def periodic_trigger():

    from .models import DateFieldTrigger

    for trigger in DateFieldTrigger.objects.filter(enable=True):
        for instance in trigger.filter.queryset.all():
            td = trigger.triggerdate(instance)
            if (
                td is not None
                and timezone.now() <= td < timezone.now() + TRIGGER_PERIOD
            ):
                run_action.apply_async(
                    (trigger.action.pk, instance._meta.label, instance.pk)
                )
